#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-, :pep:`612`-, and :pep:`646`-compliant **type parameter
utilities** (i.e., low-level callables generically handling :pep:`484`-compliant
type variables, pep:`612`-compliant parameter specifications, and
:pep:`646`-compliant type variable tuples).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep484612646Exception
from beartype._cave._cavefast import TypeParamTypes
from beartype._data.hint.datahintpep import (
    # Hint,
    TypeIs,
)
from beartype._data.hint.datahinttyping import TypeParam

# ....................{ TESTERS                            }....................
#FIXME: Unit test us up, please.
def is_hint_pep484612646_typeparam(hint: object) -> TypeIs[TypeParam]:
    '''
    :data:`True` only if the passed type hint is a **type parameter** (i.e.,
    :pep:`484`-compliant type variable, pep:`612`-compliant parameter
    specification, or :pep:`646`-compliant type variable tuples).

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this hint is a type parameter.
    '''

    # Return true only if this hint is a type parameter.
    return isinstance(hint, TypeParamTypes)

# ....................{ GETTERS                            }....................
def get_hint_pep484612646_typeparam_name(
    # Mandatory parameters.
    hint: TypeParam,

    # Optional parameters.
    exception_prefix: str = '',
) -> str:
    '''
    Unqualified basename of the passed **type parameter** (i.e.,
    :pep:`484`-compliant type variable, pep:`612`-compliant parameter
    specification, or :pep:`646`-compliant type variable tuples).

    Parameters
    ----------
    hint : TypeParam
        Type parameter to be inspected.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    str
        Unqualified basename of this type parameter.

    Raises
    ------
    BeartypeDecorHintPep484612646Exception
        If this object is *not* a type parameter.
    '''

    # If this hint is *NOT* a type parameter, raise an exception.
    if not is_hint_pep484612646_typeparam(hint):
        assert isinstance(exception_prefix, str), (
            f'{exception_prefix} not string.')
        raise BeartypeDecorHintPep484612646Exception(
            f'{exception_prefix}type hint {repr(hint)} not type parameter '
            f'(i.e., '
            f'PEP 484 "typing.TypeVar" type variable, '
            f'PEP 612 "typing.ParamSpec" parameter specification, or '
            f'PEP 646 "typing.TypeVarTuple" type variable tuple).'
        )
    # Else, this hint is a type parameter.

    # Return the name of this type parameter. Thankfully, *ALL* type parameters
    # generically conform to this simplistic API. We give minor praise.
    return hint.__name__  # type: ignore[union-attr]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **type hint reducers** (i.e., low-level callables converting type
hints from one format into another, either losslessly or in a lossy manner).

Type hint reductions imposed by this submodule are purely internal to
:mod:`beartype` itself and thus transient in nature. These reductions are *not*
permanently applied to the ``__annotations__`` dunder dictionaries of the
classes and callables annotated by these type hints.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.meta import URL_ISSUES
from beartype.roar import BeartypeDecorHintRecursionException
from beartype.typing import Optional
from beartype._cave._cavemap import NoneTypeOr
from beartype._check.convert._reduce._redmap import (
    HINT_SIGN_TO_REDUCE_HINT_CACHED_get,
    HINT_SIGN_TO_REDUCE_HINT_UNCACHED_get,
)
from beartype._check.convert._reduce._redrecurse import (
    is_hint_recursive,
    make_hint_sane_recursable,
)
from beartype._check.metadata.metadecor import BeartypeDecorMeta
from beartype._check.metadata.hint.hintsane import (
    HINT_IGNORABLE,
    HINT_SANE_IGNORABLE,
    HintOrSane,
    HintSane,
)
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._data.hint.datahintpep import Hint
from beartype._data.hint.datahinttyping import (
    DictStrToAny,
    TypeStack,
)
from beartype._util.func.arg.utilfuncargiter import ArgKind
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none
from beartype._util.kind.map.utilmapset import remove_mapping_keys
from beartype._data.kind.datakindiota import SENTINEL

# ....................{ REDUCERS                           }....................
def reduce_hint(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    arg_kind: Optional[ArgKind] = None,
    cls_stack: TypeStack = None,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
    decor_meta: Optional[BeartypeDecorMeta] = None,
    hint_parent_sane: Optional[HintSane] = None,
    is_hint_ignorable_preserved: bool = False,
    pith_name: Optional[str] = None,
    reductions_count: int = 0,
    exception_prefix: str = '',
) -> HintSane:
    '''
    Lower-level type hint reduced (i.e., converted) from the passed higher-level
    type hint if this hint is reducible *or* this hint as is otherwise (i.e., if
    this hint is irreducible).

    This reducer *cannot* be meaningfully memoized, since multiple passed
    parameters (e.g., ``pith_name``, ``cls_stack``) are typically isolated to a
    handful of callables across the codebase currently being decorated by
    :mod:`beartype`. Memoizing this reducer would needlessly consume space and
    time. To improve efficiency, this reducer is instead implemented in terms of
    two lower-level private reducers:

    * The memoized :func:`._reduce_hint_cached` reducer, responsible for
      efficiently reducing *most* (but not all) type hints.
    * The unmemoized :func:`._reduce_hint_uncached` reducer, responsible for
      inefficiently reducing the small subset of type hints contextually
      requiring these problematic parameters.

    Parameters
    ----------
    hint : Hint
        Type hint to be possibly reduced.
    arg_kind : Optional[ArgKind]
        Either:

        * If this hint annotates a parameter of some callable, that parameter's
          **kind** (i.e., :class:`.ArgKind` enumeration member conveying the
          syntactic class of that parameter, constraining how the callable
          declaring that parameter requires that parameter to be passed).
        * Else, :data:`None`.

        Defaults to :data:`None`.
    cls_stack : TypeStack, optional
        **Type stack** (i.e., either a tuple of the one or more
        :func:`beartype.beartype`-decorated classes lexically containing the
        class variable or method annotated by this hint *or* :data:`None`).
        Defaults to :data:`None`.
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object). Defaults
        to the default beartype configuration.
    decor_meta : Optional[BeartypeDecorMeta], optional
        Either:

        * If this hint annotates a parameter or return of some callable, the
          :mod:`beartype`-specific decorator metadata describing that callable.
        * Else, :data:`None`.

        Defaults to :data:`None`.
    hint_parent_sane : Optional[HintSane], default: None
        Either:

        * If the passed hint is a **root** (i.e., top-most parent hint of a tree
          of child hints), :data:`None`.
        * Else, the passed hint is a **child** of some parent hint. In this
          case, the **sanified parent type hint metadata** (i.e., immutable and
          thus hashable object encapsulating *all* metadata previously returned
          by :mod:`beartype._check.convert.convsanify` sanifiers after
          sanitizing the possibly PEP-noncompliant parent hint of this child
          hint into a fully PEP-compliant parent hint).

        Defaults to :data:`None`.
    is_hint_ignorable_preserved : bool, default: False
        Either:

        * If the caller prefers that ignorable hints reduced to a unique
          :data:`.HintSane` object *not* equal to the standard
          :data:`.HINT_SANE_IGNORABLE` singleton but instead encapsulating the
          :data:`.HINT_IGNORABLE` type hint and unique metadata describing the
          ignored hint reduce to that :data:`.HintSane` object rather than the
          :data:`.HINT_SANE_IGNORABLE` singleton, data:`True`.
        * If the caller prefers that ignorable hints reduced to a unique
          :data:`.HintSane` object *not* equal to the standard
          :data:`.HINT_SANE_IGNORABLE` singleton be transparently reduced to the
          :data:`.HINT_SANE_IGNORABLE` singleton, data:`False`. This preference
          is substantially easier for callers to handle but also technically
          lossy, as all unique metadata associated with this reduction is lost.

        Defaults to :data:`False`, as most callers neither require nor desire
        this distinction and are thus incapable of handling ignorable hints
        reduced to unique :data:`.HintSane` objects *not* equal to the standard
        :data:`.HINT_SANE_IGNORABLE` singleton. Most callers only expect the
        :data:`.HINT_SANE_IGNORABLE` singleton.
    pith_name : Optional[str], default: None
        Either:

        * If this hint annotates a parameter of some callable, the name of that
          parameter.
        * If this hint annotates the return of some callable, ``"return"``.
        * Else, :data:`None`.

        Defaults to :data:`None`.
    reductions_count : int, default: None
        Current number of total reductions internally performed by *all* calls
        to this function rooted at this function in the current call stack,
        guarding against accidental infinite recursion between lower-level
        reducers and this higher-level function. Defaults to 0.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    HintSane
        Either:

        * If this hint is ignorable, :data:`.HINT_SANE_IGNORABLE`.
        * Else if this unignorable hint is reducible to another hint, metadata
          encapsulating this reduction.
        * Else, this unignorable hint is irreducible. In this case, metadata
          encapsulating this hint unmodified.

    Raises
    ------
    BeartypeDecorHintRecursionException
        If the number of total reductions internally performed by the current
        call to this function exceeds the maximum. This exception guards against
        accidental infinite recursion between lower-level hint-specific reducers
        internally called by this higher-level hint-agnostic reducer.
    '''

    # ....................{ PREAMBLE                       }....................
    assert isinstance(arg_kind, NoneTypeOr[ArgKind]), (
        f'{repr(arg_kind)} neither argument kind nor "None".')
    assert isinstance(cls_stack, NoneTypeOr[tuple]), (
        f'{repr(cls_stack)} neither tuple nor "None".')
    assert isinstance(conf, BeartypeConf), f'{repr(conf)} not configuration.'
    assert isinstance(decor_meta, NoneTypeOr[BeartypeDecorMeta]), (
        f'{repr(hint_parent_sane)} neither decoration metadata nor "None".')
    assert isinstance(hint_parent_sane, NoneTypeOr[HintSane]), (
        f'{repr(hint_parent_sane)} neither sanified hint metadata nor "None".')
    assert isinstance(is_hint_ignorable_preserved, bool), (
        f'{repr(is_hint_ignorable_preserved)} not boolean.')
    assert isinstance(pith_name, NoneTypeOr[str]), (
        f'{repr(pith_name)} neither string nor "None".')
    assert isinstance(reductions_count, int), (
        f'{repr(reductions_count)} not integer.')
    assert isinstance(exception_prefix, str), (
        f'{repr(exception_prefix)} not string.')

    # ....................{ LOCALS                         }....................
    # Original unreduced hint passed to this reducer, preserved so as to be
    # embedded in human-readable exception messages.
    hint_old = hint

    # Currently reduced instance of this hint.
    hint_curr: Hint = hint

    # Currently reduced instance of either this hint *OR* metadata encapsulating
    # the sanification of this hint, initialized to this unreduced hint.
    hint_or_sane_curr: HintOrSane = hint

    # Previously reduced instance of either this hint *OR* metadata
    # encapsulating this hint, initialized to this unreduced hint.
    hint_or_sane_prev: HintOrSane = hint

    # Delete the passed "hint" parameter for safety. Permitting this parameter
    # to exist would only promote subtle lexical issues below, where the local
    # variable "hint_curr" is strongly preferred for disambiguity.
    del hint

    # ....................{ SEARCH                         }....................
    # Repeatedly reduce this hint to increasingly irreducible hints until this
    # hint is no longer reducible. This algorithm iteratively reduces this hint
    # with a battery of increasingly non-trivial reductions. For efficiency,
    # reductions are intentionally ordered from most to least efficient.
    while True:
        # ....................{ REDUCE                     }....................
        # For each lower-level reducer...
        for hint_reducer in _HINT_REDUCERS:
            # print(f'[reduce_hint] Reducing {hint_curr} with parent {hint_parent_sane} via {hint_reducer}...')

            # Either:
            # * If this reducer reduces this hint:
            #   * If this reduction produced supplementary metadata, metadata
            #     encapsulating the reduction of this hint by this reducer.
            #   * Else, the reduced hint reduced by this reducer.
            # * Else, this unreduced hint as is.
            hint_or_sane_curr = hint_reducer(
                hint=hint_curr,
                hint_parent_sane=hint_parent_sane,
                arg_kind=arg_kind,
                cls_stack=cls_stack,
                conf=conf,
                decor_meta=decor_meta,
                pith_name=pith_name,
                reductions_count=reductions_count,
                exception_prefix=exception_prefix,
            )
            # print(f'[reduce_hint] Reduced to {hint_or_sane_curr}!')

            # If this reduced hint is *NOT* this unreduced hint, this reducer
            # reduced this hint. Halt reducing by these lower-level reducers,
            # enabling the outer loop to decide whether to continue reducing.
            if hint_or_sane_curr is not hint_curr:
                # If this hint reduces to the ignorable "HINT_SANE_IGNORABLE"
                # metadata singleton, then halt reducing immediately.
                #
                # Note that this is merely an optimization avoiding unnecessary
                # iteration. Without this test, hints reduced to this ignorable
                # singleton would require an additional loop through the
                # "_HINT_REDUCERS" tuple. This test elides that iteration.
                if hint_or_sane_curr is HINT_SANE_IGNORABLE:
                    # print(f'[reduce_hint] Ignorably reduced!')
                    return HINT_SANE_IGNORABLE
                # Else, this hint is currently unignorable.

                # print(f'[reduce_hint] Incrementally reduced!')
                break
            # Else, this unreduced hint remains unmodified. Since this reducer
            # failed to reduce this hint, silently continue to the next reducer.
        # If the above iteration failed to "break", then this unreduced hint
        # remains unmodified across all lower-level reducers. This implies this
        # hint to now be irreducible. Halt reducing immediately.
        else:
            # print(f'[reduce_hint] Irreducible!')
            break
        # Else, the above iteration hit a "break". This hint was reduced by a
        # lower-level reducer above, implying that this hint *COULD* still be
        # reducible. Silently continue reducing.

        # ....................{ RESPOND                    }....................
        # Respond to the lower-level reduction performed above.

        # If reducing this hint generated supplementary metadata...
        if isinstance(hint_or_sane_curr, HintSane):
            # Extract the currently reduced hint from this metadata.
            hint_curr = hint_or_sane_curr.hint

            #FIXME: Should probably be performed down below outside this loop.
            # If this hint reduces to the ignorable "HINT_IGNORABLE"
            # singleton, then halt reducing immediately.
            #
            # Note that this is *NOT* merely an optimization avoiding
            # unnecessary iteration as above. While similar, this logic is
            # distinct from that above. This edge case arises when a reducer
            # avoids reducing to an ignorable hint to the higher-level ignorable
            # "HINT_SANE_IGNORABLE" metadata singleton but instead encapsulates
            # the lower-level "HINT_IGNORABLE" type hint singleton with a new
            # "HintSane" object providing unique metadata describing the ignored
            # type hint. That unique metadata enables parent reducers to
            # selectively decide how to handle ignorable child type hints.
            #
            # Examples include:
            # * The reduce_hint_pep484604() reducer for union type hints.
            if hint_curr is HINT_IGNORABLE:
                # Return either...
                return (
                    # If the caller requests that unique "HintSane" objects
                    # encapsulating the "HINT_IGNORABLE" singleton be preserved,
                    # do so by returning this metadata as is. Note that this is
                    # *NOT* what most callers expect and thus not the default;
                    hint_or_sane_curr
                    if is_hint_ignorable_preserved else
                    # Else, reduce this metadata to the standard ignorable
                    # "HINT_SANE_IGNORABLE" singleton describing ignorable
                    # hints. Note that this is what *MOST* callers expect and
                    # thus the default.
                    HINT_SANE_IGNORABLE
                )
            # Else, this hint does *NOT* reduce to the ignorable
            # "HINT_IGNORABLE" singleton. Ergo, this hint is unignorable.

            # Replace the sanified type hint metadata of the parent hint of this
            # hint by the sanified type hint metadata of this hint itself. Doing
            # so ensures that the next reducer passed the "hint_parent_sane"
            # parameter preserves this metadata during its reduction. Since the
            # most recent reducer call received the prior "hint_parent_sane"
            # parameter, that reducer has already safely preserved the parent
            # metadata by compositing that metadata into this
            # "hint_or_sane_curr" metadata that that reducer returned. Srsly.
            hint_parent_sane = hint_or_sane_curr
        # Else, reducing this hint did *NOT* generate supplementary metadata,
        # implying "hint_or_sane_curr" to be the currently reduced hint. In this
        # case, record this currently reduced hint.
        else:
            hint_curr = hint_or_sane_curr

        #FIXME: Should probably be performed above the prior "if" conditional.
        #FIXME: Currently unused, but useful. Could be required at some point.
        # If this currently reduced hint is exactly the previously reduced hint,
        # the above reducers failed to reduce this hint. Halt reducing entirely.
        #
        # Note that this is a rare (albeit valid) edge case that arises for
        # reducers that unconditionally create and return new... The above
        # "else:" block of the above "for hint_reducer in _HINT_REDUCERS:" loop
        # if hint_or_sane_curr == hint_or_sane_prev:
        #     break

        # ....................{ RECURSION                  }....................
        # Guard against infinite recursion in lower-level reductions with
        # human-readable exceptions.

        # Increment the current number of total reductions internally performed
        # by this call *BEFORE* detecting accidental recursion below.
        reductions_count += 1

        #FIXME: Unit test this, please. No idea how yet. I sigh. *sigh*
        # If the current number of total reductions internally performed
        # by this call exceeds the maximum, raise an exception.
        #
        # Note that this should *NEVER* happen, but probably nonetheless will.
        if reductions_count >= _REDUCTIONS_COUNT_MAX:  # pragma: no cover
            raise BeartypeDecorHintRecursionException(
                f'{exception_prefix}type hint {repr(hint_old)} irreducible. '
                f'Recursion detected when reducing between reduced type hints '
                f'{repr(hint_or_sane_curr)} and {repr(hint_or_sane_prev)}. '
                f'Please submit this exception traceback as a new issue '
                f'to our friendly issue tracker:\n'
                f'\t{URL_ISSUES}\n'
                f'Beartype thanks you for your noble (yet ultimately tragic) '
                f'sacrifice.'
            )
        # Else, the current number of total reductions internally performed
        # by this call is still less than the maximum. In this case, continue.

        # ....................{ PREPARE                    }....................
        # Prepare for the next iterative reduction of this "while" loop.

        # Previously reduced instance of this hint.
        hint_or_sane_prev = hint_or_sane_curr

        # Currently reduced instance of this hint, reverting back to the
        # currently visited hint in preparation for subsequent reduction.
        hint_or_sane_curr = hint_curr

    # ....................{ RETURN                         }....................
    # If this hint is *NOT* already sanified type hint metadata, this hint is
    # unignorable. Why? Because, if this hint were ignorable, this hint would
    # have been reduced to the "HINT_SANE_IGNORABLE" singleton. In this case...
    if not isinstance(hint_or_sane_curr, HintSane):
        # Encapsulate this hint with such metadata, defined as either...
        hint_or_sane_curr = (
            # If this hint has *NO* parent, this is a root hint. In this case,
            # the trivial metadata shallowly encapsulating this root hint;
            HintSane(hint_or_sane_curr)
            if hint_parent_sane is None else
            # Else, this hint has a parent. In this case, the non-trivial
            # metadata deeply encapsulating both this non-root hint *AND* all
            # metadata already associated with this parent hint.
            hint_parent_sane.permute_sane(hint=hint_or_sane_curr)
        )
    # Else, this hint is already sanified type hint metadata. In this case,
    # preserve this metadata as is.

    # Return this possibly reduced hint.
    return hint_or_sane_curr


def reduce_hint_child(
    hint: Hint, kwargs: DictStrToAny) -> HintSane:
    '''
    Lower-level child type hint reduced (i.e., converted) from the passed
    higher-level child type hint if reducible *or* this child type hint as
    is otherwise (i.e., if this child type hint is irreducible).

    This reducer is a convenience wrapper for the more general-purpose
    :func:`.reduce_hint` reducer, simplifying calls to that reducer when passed
    child hints.

    Parameters
    ----------
    hint : Hint
        Child type hint to be reduced.
    kwargs : DictStrToAny
        Keyword parameters to be passed after being unpacked to the lower-level
        :func:`.reduce_hint` reducer. For safety, this reducer silently ignores
        keyword parameters inapplicable to child hints. This includes:

        * ``arg_kind``, applicable *only* to root hints directly annotating
          callable parameters.
        * ``decor_meta``, applicable *only* to root hints directly annotating
          callable parameters or returns.
        * ``pith_name``, applicable *only* to root hints directly annotating
          callable parameters or returns.

    Returns
    -------
    HintSane
        Either:

        * If this hint is ignorable, :data:`.HINT_SANE_IGNORABLE`.
        * Else if this unignorable hint is reducible to another hint, metadata
          encapsulating this reduction.
        * Else, this unignorable hint is irreducible. In this case, metadata
          encapsulating this hint unmodified.
    '''

    # Remove all unsafe keyword parameters (i.e., parameters that are
    # inapplicable to child hints and thus *NOT* safely passable to the
    # subsequently called reduce_hint() function) from this dictionary.
    remove_mapping_keys(kwargs, _REDUCE_HINT_CHILD_ARG_NAMES_UNSAFE)

    # Return this child hint possibly reduced to a lower-level hint.
    return reduce_hint(hint=hint, **kwargs)

# ....................{ PRIVATE ~ reducers                 }....................
def _reduce_hint_overrides(
    hint: Hint,
    conf: BeartypeConf,
    hint_parent_sane: Optional[HintSane],
    **kwargs
) -> HintOrSane:
    '''
    Lower-level type hint reduced (i.e., converted) from the passed higher-level
    type hint if this hint is reducible as a **hint override** (i.e., key of
    the :attr:`.BeartypeConf.hint_overrides` dictionary of the passed beartype
    configuration) *or* this hint as is otherwise (i.e., if this hint is *not*
    as a hint override).

    Parameters
    ----------
    hint : Hint
        Type hint to be possibly reduced.
    conf : BeartypeConf
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object).
    hint_parent_sane : Optional[HintSane]
        Either:

        * If the passed hint is a **root** (i.e., top-most parent hint of a tree
          of child hints), :data:`None`.
        * Else, the passed hint is a **child** of some parent hint. In this
          case, the **sanified parent type hint metadata** (i.e., immutable and
          thus hashable object encapsulating *all* metadata previously returned
          by :mod:`beartype._check.convert.convsanify` sanifiers after
          sanitizing the possibly PEP-noncompliant parent hint of this child
          hint into a fully PEP-compliant parent hint).

    All remaining keyword parameters are silently ignored.

    Returns
    -------
    HintOrSane
        Either:

        * If this hint has already been overridden by a prior recursive
          reduction in the current tree of reductions, this hint before being
          overridden yet again (i.e., this hint as is).
        * Else if this unignorable hint is overridden by another hint, metadata
          encapsulating this override.
        * Else, this unignorable hint is *not* overridden by another hint, this
          hint as is.
    '''

    # Overridden hint to be returned, defaulting to the passed un-overridden
    # hint for safety and simplicity.
    hint_or_sane: HintOrSane = hint

    # Attempt to...
    #
    # Note that the is_object_hashable() tester is internally implemented with
    # the same Easier to Ask for Permission than Forgiveness (EAFP)-based
    # "try-except" block and is thus equally inefficient. In fact, the current
    # approach avoids an extraneous call to that tester and is thus marginally
    # faster. (Emphasis on "marginally.")
    try:
        # Hint overriding this hint if this configuration overrides this hint
        # *OR* the sentinel otherwise (i.e., if this hist is *NOT* overridden).
        #
        # Note that this raises "TypeError" when this hint is unhashable.
        # print(f'Overriding hint {repr(hint)} via {repr(conf.hint_overrides)}...')
        hint_overridden = conf.hint_overrides.get(hint, SENTINEL)

        # If neither...
        if not (
            # This hint is not overridden *NOR*...
            hint_overridden is SENTINEL or
            # Else, this hint is overridden.
            #
            # If this overridden hint is recursive, this hint has already been
            # overridden by a previously performed reduction. Avoid attempting
            # to reoverride this hint again with the same hint override; doing
            # so would provoke infinite recursion. Instead, preserve this
            # un-overridden hint by returning this hint as is.
            #
            # Certainly, various approaches to type-checking recursive hints
            # exists. @beartype currently embraces the easiest, fastest, and
            # laziest approach: just ignore all recursion! \o/
            #
            # Note that:
            # * This tester raises "TypeError" when this hint is unhashable.
            # * This tester intentionally accepts the default value "0" for the
            #   optional parameter "hint_recursable_depth_max", ensuring this
            #   overridden hint is considered to be recursive when this
            #   overridden hint has already been overridden a single time.
            #   Unlike comparable kinds of recursable hints (e.g., PEP
            #   695-compliant type aliases), hint overrides typically convey
            #   *NO* internal structure and thus merit *NO* deeper recursion.
            #   Hint overrides instruct @beartype to perform simple
            #   global-search-and-replacements on exactly matching type hints.
            #   Recursion is neither desirable nor necessary.
            #
            #   Consider the prototypical hint overrides of
            #   "BeartypeConf(hint_overrides={float: float | int})". After
            #   expanding the builtin "float" type to the PEP 604-compliant
            #   union "float | int", attempting to recursively re-apply the same
            #   expansion silently reduces to a noop (e.g., "float | int"
            #   expands to "float | float | int", equal to "float | int").
            is_hint_recursive(hint=hint, hint_parent_sane=hint_parent_sane)
        ):
            # Then this overridden hint is *NOT* recursive, implying this hint
            # *CANNOT* have already been overridden by a previously performed
            # reduction. Why? Because an overridden hint revisited by the
            # current breadth-first search would by definition by recursive.

            # Metadata guarding this hint against infinite recursion, recording
            # this hint as already having been overridden *BEFORE* reducing and
            # thus forgetting this hint.
            #
            # Note that this intentionally replaces the metadata encapsulating
            # the sanification of the parent hint of this hint by the metadata
            # encapsulating the sanification of this hint itself. Doing so
            # ensures that the next reducer passed the "hint_parent_sane"
            # parameter preserves this metadata during its reduction. Since the
            # most recent reducer call received the prior "hint_parent_sane"
            # parameter, that reducer has already safely preserved the parent
            # metadata by compositing that metadata into this
            # "hint_or_sane_curr" metadata that that reducer returned. Srsly.
            hint_or_sane = make_hint_sane_recursable(
                # The recursable form of this overridden hint is the
                # pre-overridden hint tested above by the is_hint_recursive()
                # recursion guard.
                hint_recursable=hint,
                # The non-recursable form of this overridden hint is the
                # overridden hint encapsulated by the metadata returned by this
                # factory.
                hint_nonrecursable=hint_overridden,
                hint_parent_sane=hint_parent_sane,
            )
        # Else, this overridden hint is recursive. In this case, preserve this
        # un-overridden hint rather than reducing this hint to the ignorable
        # "HINT_SANE_IGNORABLE" singleton. Why? Because un-overridden hints are
        # themselves valid type hints and thus have semantic meaning in and of
        # themselves (e.g., the "float" in the hint override
        # "BeartypeConf(hint_overrides={float: float | int})" has semantic
        # meaning as a builtin type).
    # If doing so raises a "TypeError", this hint is unhashable and thus
    # inapplicable for hint overriding. In this case, preserve this hint as is.
    except TypeError:
        pass

    # Return this possibly overridden hint.
    return hint_or_sane


def _reduce_hint_cached(
    hint: Hint, exception_prefix: str, **kwargs) -> HintOrSane:
    '''
    Lower-level type hint reduced (i.e., converted) from the passed higher-level
    type hint if this hint is reducible by a **memoized reducer** (i.e.,
    lower-level reducer accepting *only* a passed hint and thus readily amenable
    to memoization) *or* this hint as is otherwise (i.e., if this hint is *not*
    reducible by a memoized reducer).

    Parameters
    ----------
    hint : Hint
        Type hint to be possibly reduced.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    All remaining keyword parameters are silently ignored.

    Returns
    -------
    HintOrSane
        Either:

        * If this hint is ignorable, :data:`.HINT_SANE_IGNORABLE`.
        * Else if this unignorable hint is reducible to another hint by a
          memoized reducer, metadata encapsulating this reduction.
        * Else, this hint unmodified as is.
    '''

    # Reduced hint to be returned, defaulting to the passed unreduced hint.
    hint_or_sane: HintOrSane = hint

    # Sign uniquely identifying this hint if this hint is PEP-compliant *OR*
    # "None" otherwise (e.g., if this hint is PEP-noncompliant).
    hint_sign = get_hint_pep_sign_or_none(hint)

    # Memoized reducer reducing this hint if any *OR* "None" otherwise.
    hint_reducer_cached = HINT_SIGN_TO_REDUCE_HINT_CACHED_get(hint_sign)

    # If a memoized reducer reduces this hint...
    if hint_reducer_cached is not None:
        # print(f'[_reduce_hint_cached] Reducing cached hint {repr(hint)}...')

        #FIXME: [SPEED] Is there any point to passing the "exception_prefix"
        #parameter? Possibly. Not sure. Isn't this parameter a constant? No?
        #Does it actually vary with context? Can't recall. Investigate up!

        # Reduce this hint by calling this reducer.
        #
        # Note that parameters are intentionally passed positionally to this
        # possibly memoized callable prohibiting keyword parameters.
        hint_or_sane = hint_reducer_cached(hint, exception_prefix)
    # Else, *NO* memoized reducer reduces this hint. In this case, preserve this
    # hint as is.

    # Return this possibly reduced hint.
    return hint_or_sane


def _reduce_hint_uncached(hint: Hint, **kwargs) -> HintOrSane:
    '''
    Lower-level type hint reduced (i.e., converted) from the passed higher-level
    type hint if this hint is reducible by a **unmemoized reducer** (i.e.,
    lower-level reducer accepting *only* a passed hint and thus readily amenable
    to memoization) *or* this hint as is otherwise (i.e., if this hint is *not*
    reducible by a unmemoized reducer).

    Parameters
    ----------
    hint : Hint
        Type hint to be possibly reduced.

    All remaining keyword parameters are silently ignored.

    Returns
    -------
    HintOrSane
        Either:

        * If this hint is ignorable, :data:`.HINT_SANE_IGNORABLE`.
        * Else if this unignorable hint is reducible to another hint by a
          unmemoized reducer, metadata encapsulating this reduction.
        * Else, this hint unmodified as is.
    '''

    # Reduced hint to be returned, defaulting to the passed unreduced hint.
    hint_or_sane: HintOrSane = hint

    # Sign uniquely identifying this hint if this hint is PEP-compliant *OR*
    # "None" otherwise (e.g., if this hint is PEP-noncompliant).
    hint_sign = get_hint_pep_sign_or_none(hint)

    # Unmemoized reducer reducing this hint if any *OR* "None" otherwise.
    hint_reducer_uncached = HINT_SIGN_TO_REDUCE_HINT_UNCACHED_get(hint_sign)

    # If a unmemoized reducer reduces this hint...
    if hint_reducer_uncached is not None:
        # print(f'[_reduce_hint_cached] Reducing cached hint {repr(hint)}...')

        # Reduce this hint by calling this reducer.
        hint_or_sane = hint_reducer_uncached(hint=hint, **kwargs)
    # Else, *NO* unmemoized reducer reduces this hint. In this case, preserve
    # this hint as is.

    # Return this possibly reduced hint.
    return hint_or_sane

# ....................{ PRIVATE ~ globals                  }....................
_HINT_REDUCERS = (
    # ....................{ PHASE ~ override               }....................
    # Attempt to reduce this hint to another hint configured by a user-defined
    # hint override *BEFORE* applying standard reductions. User preference
    # assumes precedence over standard precedent. Nice one-liner, huh? </heh>
    _reduce_hint_overrides,

    # ....................{ PHASE ~ context-free           }....................
    # Attempt to reduce this hint with a context-free reduction *BEFORE*
    # reducing this hint with a contextual reduction. Due to *NOT* depending on
    # contextual state, context-free reductions are readily memoizable and thus
    # faster than contextual reductions.
    _reduce_hint_cached,

    # ....................{ PHASE ~ contextual             }....................
    # Attempt to reduce this hint with a contextual reduction. Due to depending
    # on contextual state, contextual reductions are *NOT* be readily memoizable
    # and thus slower than context-free reductions.
    _reduce_hint_uncached,
)
'''
Tuple of all private high-level reducers defined by this submodule above.
'''


_REDUCTIONS_COUNT_MAX = 64
'''
Maximum number of total reductions internally performed by each call of the
:func:`reduce_hint` function.

This constant is a relatively arbitrary magic number selected so as to guard
against accidental infinite recursion between lower-level PEP-specific reducers
internally called by :func:`reduce_hint`.
'''


_REDUCE_HINT_CHILD_ARG_NAMES_UNSAFE = frozenset((
    # Applicable *ONLY* to root hints directly annotating callable parameters.
    'arg_kind',

    # Applicable *ONLY* to root hints directly annotating callable parameters
    # or returns.
    'decor_meta',
    'pith_name',
))
'''
Frozen set of the names of all **unsafe child type hint reducer keyword
parameters** (i.e., keyword parameters inapplicable to child type hints and thus
*not* safely passable from the higher-level :func:`.reduce_hint_child` to
lower-level :func:`.reduce_hint` reducer.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- and :pep:`585`-compliant **generic type hint
iterators** (i.e., low-level callables generically iterating over both
:pep:`484`- and :pep:`585`-compliant generic class hierarchies).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: Generalize to support generic "TypedDict" and "NamedTuple" subclasses as
#follows:
#* Improve the iterator below as follows:
#      hint_sign_nongeneric = get_hint_pep_sign_or_none(
#          hint, _IGNORE_HINT_SIGNS_PEP484585_GENERIC_UNSUBSCRIPTED)
#
#      # Note that this constitutes an *EXTREMELY* rare edge case. In fact, it
#      # took ten years for the first @beartype user to even hit this case.
#      # Ergo, efficiency is absolutely *NOT* a concern. Maintainability is.
#      if hint_sign_nongeneric is not None:
#          hint_bases_direct = (hint,) + hint_bases_direct
#* Exhaustively unit test this edge case.
#
#Pretty sure that should do it. Insanity, but surprisingly amenable to hacky.
#FIXME: *WOAH.* Okay. The above commentary absolutely *DID NOT DO IT.* The key
#insight from all of this is that, although *MOST* type hints are indeed
#uniquely identifiably by a single sign, *SOME* type hints are (sadly) only
#uniquely identifiably by two or more signs. These are the latter sort.
#
#Sure, we could try to hack around that. But those hacks will inevitably fail.
#It'd be *A LOT* saner to get out ahead of this issue by just acknowledging
#that, in the general case, only a tuple of signs suffices to identify a hint.
#But you are now cogitating: "Uhh... but doesn't the *ENTIRE* codebase assume a
#one-to-one relation between a hint and a sign?" The answer, of course, is:
#"Yes. Yes, it does."
#
#But that doesn't mean we can't have our cake and eat it, too. That just means
#we need to be careful about the way we our cake and eat it, too. Notably:
#* Define a new "beartype._util.hint.pep.utilpepsign" submodule. This submodule
#  should contain *ALL* sign-related functionality, which is rapidly becoming
#  non-trivial. It was always non-trivial, but now it's *REALLY* non-trivial.
#* In this submodule:
#  * Define a new private _get_hint_pep_sign_unique_or_none() getter, refactored
#    from the existing get_hint_pep_sign_or_none() getter by dropping all
#    reference to the "HintSignPep484585GenericUnsubscripted" sign. Otherwise,
#    the implementation should be the exact same as get_hint_pep_sign_or_none().
#  * Abandon the new optional "ignore_hint_signs" parameter passed to the
#    get_hint_pep_signs_or_none() getter. Nice idea, but ultimately flawed. We
#    can and must do *SUBSTANTIALLY* better than that chaos.
#  * Define a new get_hint_pep_signs_or_none() getter, refactored from the
#    existing get_hint_pep_sign_or_none() getter with modifications as follows:
#    @callable_cached
#    def get_hint_pep_signs_or_none(hint: Any) -> TupleHintSign:
#        hint_signs_list = acquire_instance(list)
#
#        if is_hint_pep484585_generic_unsubbed(hint):
#            hint_signs_list.append(HintSignPep484585GenericUnsubscripted)
#
#        hint_sign_unique = _get_hint_pep_sign_unique_or_none(hint)
#        hint_signs_list.append(hint_sign_unique)
#
#        hint_signs = tuple(hint_signs_list)
#        release_instance(hint_signs_list)
#        return hint_signs
#  * Refactor the existing get_hint_pep_sign_or_none() getter to trivially defer
#    to the new get_hint_pep_signs_or_none() getter as follows:
#    # Note: don't bother caching this! get_hint_pep_signs_or_none() is already
#    # memoized, which is more than enough.
#    def get_hint_pep_sign_or_none(hint: Any) -> TupleHintSign:
#        hint_signs = get_hint_pep_signs_or_none(hint)
#        return hint_signs[0]
#* *TEST THAT EVERYTHING STILL WORKS.*
#* Refactor "codemake" to call get_hint_pep_signs_or_none() instead of
#  get_hint_pep_sign_or_none() and then iterate over the result: e.g.,
#    hint_curr_signs = get_hint_pep_signs_or_none(hint_curr)
#
#    #FIXME: [SPEED] Refactor into a "while" loop for speed. whatevah!
#    for hint_curr_sign in hint_curr_signs:
#        ...
#* *TEST THAT EVERYTHING STILL WORKS.*
#* Refactor "errcause" similarly.
#* *TEST THAT EVERYTHING STILL WORKS.*
#
#Pretty cool, right? And it genuinely is. Most callers shouldn't need to care
#about this distinction. They can continue to call get_hint_pep_sign_or_none()
#and just pretend that hints have only one sign. Hopefully, the only callers
#that care are our low-level code generators. Let's pretend this is the case.
#*shudders*
#FIXME: Actually... belay that. While useful, the above is overkill for the
#moment. Instead, just call the new
#get_hint_pep484585_generic_unsubbed_sign_nongeneric_or_none() getter in
#both "codemake" and the appropriate "beartype._check.error" submodule, please.

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep484585Exception
from beartype.typing import (
    Iterable,
    Tuple,
)
from beartype._check.convert.convsanify import sanify_hint_child
from beartype._check.metadata.hint.hintsane import (
    HINT_SANE_IGNORABLE,
    HintSane,
    # IterableHintSane,
)
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._data.hint.datahinttyping import (
    TypeException,
    TypeStack,
)
from beartype._data.hint.pep.sign.datapepsigncls import HintSign
from beartype._util.cache.pool.utilcachepoollistfixed import (
    FIXED_LIST_SIZE_MEDIUM,
    acquire_fixed_list,
    release_fixed_list,
)
from beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget import (
    get_hint_pep484585_generic_base_extrinsic_sign_or_none,
    get_hint_pep484585_generic_bases_unerased,
)
from beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest import (
    is_hint_pep484585_generic_user)
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

# ....................{ ITERATORS                          }....................
#FIXME: Unit test us up, please.
def iter_hint_pep484585_generic_unsubbed_bases_unerased(
    # Mandatory parameters.
    hint_sane: HintSane,

    # Optional parameters.
    cls_stack: TypeStack = None,
    conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
    exception_cls: TypeException = BeartypeDecorHintPep484585Exception,
    exception_prefix: str = '',
# ) -> IterableHintSane:
) -> Iterable[Tuple[HintSane, HintSign]]:
    '''
    Generator iteratively yielding the one or more **unerased
    pseudo-superclasses** (i.e., unignorable PEP-compliant type hints originally
    declared as transitive superclasses prior to type erasure) of the passed
    :pep:`484`- or :pep:`585`-compliant unsubscripted generic, effectively
    performing a breadth-first search (BFS) over these pseudo-superclasses.

    This generator yields the full tree of all pseudo-superclasses by
    transitively visiting both all direct pseudo-superclasses of this generic
    *and* all indirect pseudo-superclasses transitively superclassing all direct
    pseudo-superclasses of this generic. For efficiency, this generator is
    internally implemented with an efficient imperative First In First Out
    (FILO) queue rather than an inefficient (and dangerous, due to both
    unavoidable stack exhaustion and avoidable infinite recursion) tree of
    recursive function calls.

    Note that there exist two kinds of pseudo-superclasses with respect to
    type-checking. Each pseudo-superclass yielded by this generator is either:

    * An **intrinsic pseudo-superclass** (i.e., whose type-checking is
      intrinsically defined as a type hint such that all data required to
      type-check this pseudo-superclass is fully defined by this hint). *All*
      intrinsic pseudo-superclasses are valid type hints. This is the common
      case and, indeed, almost all cases. Examples include :pep:`484`- and
      :pep:`585`-compliant subscripted container type hints: e.g.,

        .. code-block:: python

           # The PEP 585-compliant "list[T]" pseudo-superclass is a valid hint
           # whose type-checking is intrinsic to this hint.
           class GenericList[T](list[T]):
               def generic_method(self, arg: T) -> T:
                   return arg

    * An **extrinsic pseudo-superclass (i.e., whose type-checking is
      extrinsically defined by this unsubscripted generic such that only the
      combination of this pseudo-superclass and this unsubscripted generic
      suffices to provide all data required to type-check this
      pseudo-superclass). Extrinsic pseudo-superclasses are *not* necessarily
      valid type hints, though some might be. Examples include:

      * **Generic named tuples** (i.e., types subclassing both the
        :pep:`484`-compliant :class:`typing.Generic` superclass *and* the
        :pep:`484`-compliant :class:`typing.NamedTuple` superclass): e.g.,

        .. code-block:: python

           from typing import Generic, NamedTuple
           class GenericNamedTuple[T](NamedTuple, Generic[T]):
               generic_item: T

        When iterating over the :class:`typing.NamedTuple` pseudo-superclass of
        a generic typed dictionary, this generator yields the 2-tuple
        ``(hint_sane.hint, HintSignNamedTuple)`` (e.g.,
        ``(GenericNamedTuple, HintSignNamedTuple)`` for the above generic).

      * **Generic typed dictionaries** (i.e., types subclassing both the
        :pep:`484`-compliant :class:`typing.Generic` superclass *and* the
        :pep:`589`-compliant :class:`typing.TypedDict` superclass): e.g.,

        .. code-block:: python

           from typing import Generic, TypedDict
           class GenericTypedDict[T](TypedDict, Generic[T]):
               generic_item: T

        When iterating over the :class:`typing.TypedDict` pseudo-superclass of
        a generic typed dictionary, this generator yields the 2-tuple
        ``(hint_sane.hint, HintSignTypedDict)`` (e.g.,
        ``(GenericTypedDict, HintSignTypedDict)`` for the above generic).

    Motivation
    ----------
    Ideally, a BFS would *not* be necessary. Instead, pseudo-superclasses
    visited by this BFS should be visitable as is via whatever external parent
    BFS is currently iterating over the tree of all transitive type hints (e.g.,
    our code generation algorithm implemented by the
    :func:`beartype._check.code.codemake.make_func_pith_code` function).
    That's how we transitively visit all other kinds of type hints, right?
    Sadly, that simple solution fails to scale to all possible edge cases that
    arise with generics. Why? Because our code generation algorithm sensibly
    requires that *only* unignorable hints may be enqueued onto its outer BFS.
    Generics confound that constraint. Some pseudo-superclasses are
    paradoxically:

    * Ignorable from the perspective of code generation. *No* type-checking code
      should be generated for these pseudo-superclasses. See reasons below.
    * Unignorable from the perspective of algorithm visitation. These
      pseudo-superclasses generate *no* code but may themselves subclass other
      pseudo-superclasses for which type-checking code should be generated and
      which must thus be visited by our outer BFS.

    Paradoxical pseudo-superclasses include:

    * User-defined :pep:`484`-compliant subgenerics (i.e., user-defined generics
      subclassing one or more parent user-defined generic superclasses).
    * User-defined :pep:`544`-compliant subprotocols (i.e., user-defined
      protocols subclassing one or more parent user-defined protocol
      superclasses).

    Consider this example :pep:`544`-compliant subprotocol:

    .. code-block:: pycon

       >>> import typing as t
       >>> class UserProtocol(t.Protocol[t.AnyStr]): pass
       >>> class UserSubprotocol(UserProtocol[str], t.Protocol): pass
       >>> UserSubprotocol.__orig_bases__
       (UserProtocol[str], typing.Protocol)  # <-- good
       >>> UserProtocolUnerased = UserSubprotocol.__orig_bases__[0]
       >>> UserProtocolUnerased is UserProtocol
       False
       >>> isinstance(UserProtocolUnerased, type)
       False  # <-- bad

    :pep:`585`-compliant generics suffer no such issues:

    .. code-block:: pycon

       >>> from beartype._util.hint.pep.proposal.pep585 import is_hint_pep585_builtin_subbed
       >>> class UserGeneric(list[int]): pass
       >>> class UserSubgeneric(UserGeneric[int]): pass
       >>> UserSubgeneric.__orig_bases__
       (UserGeneric[int],)
       >>> UserGenericUnerased = UserSubgeneric.__orig_bases__[0]
       >>> isinstance(UserGenericUnerased, type)
       True  # <-- good
       >>> UserGenericUnerased.__mro__
       (UserGeneric, list, object)
       >>> is_hint_pep585_builtin_subbed(UserGenericUnerased)
       True

    Iteratively walking up the unerased inheritance hierarchy for any such
    paradoxical generic or protocol subclass (e.g., ``UserSubprotocol`` but
    *not* ``UserSubgeneric`` above) would visit a user-defined generic or
    protocol pseudo-superclass subscripted by type variables. Due to poorly
    defined obscurities in the :mod:`typing` implementation, that
    pseudo-superclass is *not* actually a class but rather an instance of a
    private :mod:`typing` class (e.g., :class:`typing._SpecialForm`). This
    algorithm would then detect that pseudo-superclass as neither a generic nor
    a :mod:`typing` object and thus raise an exception. Fortunately, that
    pseudo-superclass conveys no meaningful intrinsic semantics with respect to
    type-checking; its only use is to register its own pseudo-superclasses (one
    or more of which could convey meaningful intrinsic semantics with respect to
    type-checking) for visitation by this BFS.

    Parameters
    ----------
    hint_sane : HintSane
        **Sanified type hint metadata** (i.e., :data:`.HintSane` object)
        encapsulating the :pep:`484`- or :pep:`585`-compliant unsubscripted
        generic to be inspected.
    conf : BeartypeConf, optional
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for the passed object). Defaults
        to :data:`.BEARTYPE_CONF_DEFAULT`, the default :math:`O(1)`
        type-checking configuration.
    cls_stack : TypeStack, optional
        **Type stack** (i.e., either a tuple of the one or more
        :func:`beartype.beartype`-decorated classes lexically containing the
        class variable or method annotated by this hint *or* :data:`None`).
        Defaults to :data:`None`.
    exception_cls : TypeException
        Type of exception to be raised. Defaults to
        :exc:`BeartypeDecorHintPep484585Exception`.
    exception_prefix : str, optional
        Human-readable substring prefixing the representation of this object in
        the exception message. Defaults to the empty string.

    Returns
    -------
    Iterable[Tuple[HintSane, HintSign]]
        Generator iteratively yielding one or more 2-tuples ``(hint_sane,
        hint_sign)``, where:

        * ``hint_sane`` is an unignorable unerased transitive pseudo-superclass
          originally declared as a superclass prior to its type erasure of this
          unsubscripted generic.
        * ``hint_sign`` is the sign uniquely identifying this pseudo-superclass.
          Since this sign does *not* necessarily correspond to the sign returned
          by the :func:`.get_hint_pep_sign_or_none` getter when passed this
          pseudo-superclass, callers should take care to preserve this sign.

    Raises
    ------
    exception_cls
        If this hint is *not* a generic.

    See Also
    --------
    :func:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget.get_hint_pep484585_generic_type_or_none`
        Further details.
    '''
    # assert is_hint_pep484585_generic_unsubbed(hint_sane.hint)
    assert isinstance(hint_sane, HintSane), (
        f'{repr(hint_sane)} not sanified metadata.')

    # ....................{ LOCALS ~ bases : direct        }....................
    # This unsubscripted generic.
    hint = hint_sane.hint

    # Tuple of the one or more unerased pseudo-superclasses originally listed as
    # superclasses prior to their type erasure by this unsubscripted generic.
    hint_bases_direct = get_hint_pep484585_generic_bases_unerased(
        hint=hint,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )
    # print(f'generic {hint} hint_bases_direct: {hint_bases_direct}')

    # ....................{ LOCALS ~ bases : indirect      }....................
    # Fixed list of the one or more unerased transitive pseudo-superclasses
    # originally listed as superclasses prior to their type erasure by this
    # generic that have yet to be visited by the breadth-first search (BFS) over
    # these pseudo-superclasses performed below.
    hint_bases = acquire_fixed_list(size=FIXED_LIST_SIZE_MEDIUM)

    # 0-based index of the currently visited pseudo-superclass of this list.
    hint_bases_index_curr = 0

    # 0-based index of one *PAST* the last pseudo-superclass of this list.
    hint_bases_index_past_last = len(hint_bases_direct)

    # Initialize this list to these direct pseudo-superclasses of this generic.
    hint_bases[0:hint_bases_index_past_last] = hint_bases_direct
    # print(f'generic pseudo-superclasses [initial]: {repr(hint_bases_direct}')

    # ....................{ SEARCH                         }....................
    # While the 0-based index of the next visited pseudo-superclass does *NOT*
    # exceed that of the last pseudo-superclass in this list, there remains one
    # or more pseudo-superclasses to be visited in this BFS. Let us do so.
    #
    # Each iteration of this search is subdivided into two phases, enabling this
    # search to discern between intrinsic and extrinsic pseudo-superclasses --
    # whose handling is fundamentally different. See the function docstring for
    # the distinction between the two.
    while hint_bases_index_curr < hint_bases_index_past_last:
        # ....................{ LOCALS                     }....................
        # Current pseudo-superclass of this unsubscripted generic.
        hint_base = hint_bases[hint_bases_index_curr]

        # ....................{ PHASE ~ extrinsic          }....................
        # In this first phase, we:
        # 1. Decide whether this pseudo-superclass is extrinsic.
        # 2. If so, return this subscripted generic (rather than this
        #    pseudo-superclass) *AND* the sign uniquely identifying this
        #    pseudo-superclass. To distinguish this sign from the normal sign
        #    identifying a hint, this sign is referred to as the "subsign"
        #    (i.e., subclass sign applicable to this subscripted generic
        #    subclassing this pseudo-superclass rather than this
        #    pseudo-superclass itself).
        #
        # Note that:
        # * Extrinsic pseudo-superclasses are *EXTREMELY* rare. Almost all
        #   pseudo-superclasses are intrinsic.
        # * Extrinsic pseudo-superclasses are efficiently detectable in
        #   non-amortized O(1) time. Even though extrinsic pseudo-superclasses
        #   are rare, the cost of handling them is thankfully minimal.
        # * Extrinsic pseudo-superclasses *MUST* be detected before intrinsic
        #   pseudo-superclasses. Some extrinsic pseudo-superclasses (e.g.,
        #   "typing.TypedDict") are also valid type hints and thus also valid
        #   intrinsic pseudo-superclasses. Extrinsic pseudo-superclasses convey
        #   more fine-grained data for type-checking purposes than intrinsic
        #   pseudo-superclasses; the former are thus preferable to the latter.

        # Sign additionally identifying this pseudo-superclass if this
        # pseudo-superclass is extrinsic *OR* "None" otherwise (i.e., if this
        # pseudo-superclass is intrinsic).
        hint_base_extrinsic_sign = (
            get_hint_pep484585_generic_base_extrinsic_sign_or_none(
                hint_base=hint_base,
                exception_cls=exception_cls,
                exception_prefix=exception_prefix,
            ))

        # If this pseudo-superclass is extrinsic, yield the 2-tuple of:
        # * This *UNSUBSCRIPTED GENERIC* (rather than this extrinsic
        #   pseudo-superclass). Why? Because, by definition, an extrinsic
        #   pseudo-superclass itself conveys insufficient metadata required
        #   to type-check the unsubscripted generic subclassing that
        #   pseudo-superclass; that metadata is extrinsic to that
        #   pseudo-superclass, residing in this unsubscripted generic.
        # * This sign.
        if hint_base_extrinsic_sign is not None:
            yield hint_sane, hint_base_extrinsic_sign
        # Else, this pseudo-superclass is intrinsic. Continue to the next phase
        # deciding the sign identifying this intrinsic.
        #
        # ....................{ PHASE ~ intrinsic          }....................
        # In this second phase, we fallback to treating this pseudo-superclass
        # as intrinsic by returning this this pseudo-superclass *AND* the sign
        # uniquely identifying this pseudo-superclass. This is the common case.
        else:
            # Sane pseudo-superclass sanified from this possibly insane
            # pseudo-superclass if sanifying this pseudo-superclass did not
            # generate supplementary metadata *OR* that metadata (i.e., if doing
            # so generated supplementary metadata).
            hint_base_sane = sanify_hint_child(
                hint=hint_base,
                hint_parent_sane=hint_sane,
                cls_stack=cls_stack,
                conf=conf,
                exception_prefix=exception_prefix,
            )
            # print(f'generic {hint} base: {repr(hint_base)}')

            # If this pseudo-superclass is unignorable...
            if hint_base_sane is not HINT_SANE_IGNORABLE:
                # Sanified pseudo-superclass encapsulated by this metadata.
                hint_base = hint_base_sane.hint

                # If this pseudo-superclass is a user-defined PEP 484-compliant
                # generic or 544-compliant protocol, generate *NO* type-checking
                # code for this pseudo-superclass; instead, we only enqueue
                # *ALL* parent pseudo-superclasses of this child
                # pseudo-superclass for visitation by later iteration of this
                # inner BFS.
                #
                # See "hints_bases" for explanatory commentary.
                if is_hint_pep484585_generic_user(hint_base):
                    # Tuple of the one or more parent pseudo-superclasses of
                    # this child pseudo-superclass.
                    hint_base_bases = get_hint_pep484585_generic_bases_unerased(
                        hint=hint_base,
                        exception_cls=exception_cls,
                        exception_prefix=exception_prefix,
                    )

                    # 0-based index of the last pseudo-superclass of this list
                    # *BEFORE* adding onto this list.
                    hint_bases_index_past_last_prev = hint_bases_index_past_last

                    # 0-based index of the last pseudo-superclass of this list
                    # *AFTER* adding onto this list.
                    hint_bases_index_past_last += len(hint_base_bases)

                    # Enqueue these superclasses onto this list.
                    hint_bases[
                        hint_bases_index_past_last_prev:
                        hint_bases_index_past_last
                    ] = hint_base_bases
                # Else, this pseudo-superclass is neither an ignorable
                # user-defined PEP 484-compliant generic *NOR* an ignorable
                # 544-compliant protocol. In either case, this pseudo-superclass
                # is unignorable. In this case...
                else:
                    # Sign uniquely identifying this intrinsic pseudo-superclass
                    # if this pseudo-superclass is PEP-compliant *OR* "None"
                    # otherwise (i.e., if this pseudo-superclass is
                    # PEP-noncompliant).
                    hint_base_intrinsic_sign = get_hint_pep_sign_or_none(hint_base)

                    # If this pseudo-superclass is PEP-compliant, this
                    # pseudo-superclass is a type hint conveying meaningful
                    # semantics. Yield the 2-tuple of:
                    # * This intrinsic pseudo-superclass.
                    # * This sign.
                    if hint_base_intrinsic_sign is not None:
                        yield hint_base_sane, hint_base_intrinsic_sign
                    # Else, this pseudo-superclass is an isinstanceable type
                    # conveying *NO* meaningful semantics and is thus
                    # effectively ignorable. Why? Because the caller already
                    # type-checks this pith against the generic subclassing this
                    # superclass and thus this superclass as well in an
                    # isinstance() call (e.g., in the
                    # "CODE_PEP484585_GENERIC_PREFIX" snippet leveraged by the
                    # "beartype._check.code.codemake" submodule).
            # Else, this pseudo-superclass is ignorable.
            # else:
            #     print(f'Ignoring generic {repr(hint)} base {repr(hint_base)}...')
            #     print(f'Is generic {hint} base {repr(hint_base)} type? {isinstance(hint_base, type)}')

        # Nullify the previously visited pseudo-superclass in this list.
        hint_bases[hint_bases_index_curr] = None

        # Increment the 0-based index of the next visited pseudo-superclass in
        # this list *BEFORE* visiting that pseudo-superclass but *AFTER*
        # performing all other logic for the current pseudo-superclass.
        hint_bases_index_curr += 1

    # ....................{ POSTAMBLE                      }....................
    # Release this list. Pray for salvation, for we find none here.
    release_fixed_list(hint_bases)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-compliant **type variable reducers** (i.e., low-level
callables converting :pep:`484`-compliant type variables to lower-level type
hints more readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import (
    BeartypeDecorHintPep484TypeVarException,
    BeartypeDecorHintPep484TypeVarViolation,
)
from beartype.typing import (
    Optional,
    TypeVar,
)
from beartype._check.convert._reduce._redrecurse import (
    is_hint_recursive,
    make_hint_sane_recursable,
)
from beartype._check.metadata.hint.hintsane import (
    HINT_SANE_IGNORABLE,
    HINT_SANE_RECURSIVE,
    HintOrSane,
    HintSane,
)
from beartype._data.hint.datahintpep import (
    Hint,
    TupleHints,
    TypeVarToHint,
)
from beartype._data.hint.datahinttyping import (
    TupleTypeVars,
)
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.cls.pep.clspep3119 import is_object_issubclassable
from beartype._util.hint.nonpep.utilnonpeptest import is_hint_nonpep_type
from beartype._util.hint.pep.proposal.pep484.pep484typevar import (
    get_hint_pep484_typevar_bound_or_none,
    is_hint_pep484_typevar,
)
from beartype._util.hint.pep.utilpepget import (
    get_hint_pep_args,
    get_hint_pep_origin,
    get_hint_pep_typevars,
)
from beartype._util.hint.pep.utilpeptest import is_hint_pep
from beartype._util.kind.map.utilmapfrozen import FrozenDict

# ....................{ REDUCERS                           }....................
def reduce_hint_pep484_typevar(
    hint: Hint,
    hint_parent_sane: Optional[HintSane],
    exception_prefix: str,
    **kwargs
) -> HintSane:
    '''
    Reduce the passed :pep:`484`-compliant **type variable** (i.e.,
    :class:`typing.TypedDict` instance) to a lower-level type hint currently
    supported by :mod:`beartype`.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : Hint
        Type variable to be reduced.
    hint_parent_sane : Optional[HintSane]
        Either:

        * If the passed hint is a **root** (i.e., top-most parent hint of a tree
          of child hints), :data:`None`.
        * Else, the passed hint is a **child** of some parent hint. In this
          case, the **sanified parent type hint metadata** (i.e., immutable and
          thus hashable object encapsulating *all* metadata previously returned
          by :mod:`beartype._check.convert.convsanify` sanifiers after
          sanitizing the possibly PEP-noncompliant parent hint of this child
          hint into a fully PEP-compliant parent hint).
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    All remaining passed arguments are silently ignored.

    Returns
    -------
    HintSane
        Either:

        * If the type variable lookup table encapsulated by the passed sanified
          parent type hint metadata maps this type variable to another type
          hint, the latter.
        * Else if this type variable is both unbounded and unconstrained, this
          type variable is ignorable. In this case, the
          :data:`.HINT_SANE_IGNORABLE` singleton.
        * Else, this type variable's lover-level bounds or constraints.
    '''
    # print(f'Reducing PEP 484 type variable {hint} with parent hint {hint_parent_sane}...')

    # ....................{ RECURSE                        }....................
    # If this subscripted hint is recursive, ignore this subscripted hint to
    # avoid infinite recursion.
    if is_hint_recursive(
        hint=hint,
        hint_parent_sane=hint_parent_sane,
        hint_recursable_depth_max=_HINT_PEP484_TYPEVAR_RECURSABLE_DEPTH_MAX,
    ):
        # print(f'Ignoring recursive type variable {hint} with parent {hint_parent_sane}!')
        return HINT_SANE_RECURSIVE
    # Else, this subscripted hint is *NOT* recursive.

    # ....................{ PHASE                          }....................
    # This reducer is divided into two phases:
    # 1. The first phase decides the type variable lookup table for this alias.
    # 2. The second phase decides the recursion guard for this alias.
    #
    # Both phases are non-trivial. The output of each phase is sanified hint
    # metadata (i.e., a "HintSane" object) containing the result of the decision
    # problem decided by that phase.

    # Reduced hint to be returned, defaulting to this type variable.
    hint_reduced = hint

    # ....................{ PHASE ~ 1 : lookup             }....................
    # If...
    if (
        # This type variable is not a root hint and thus has a parent hint
        # *AND*...
        hint_parent_sane is not None and
        # A parent hint of this type variable maps one or more type variables...
        hint_parent_sane.typevar_to_hint
    ):
        # Type variable lookup table of this parent hint, localized for
        # usability and negligible efficiency.
        typevar_to_hint = hint_parent_sane.typevar_to_hint

        # If a parent hint of this type variable maps exactly one type variable,
        # prefer a dramatically faster and simpler approach.
        if len(typevar_to_hint) == 1:
            # Hint mapped to by this type variable if one or more parent hints
            # previously mapped this type variable to a hint *OR* this hint as
            # is otherwise (i.e., if this type variable is unmapped).
            #
            # Note that this one-liner looks ridiculous, but actually works.
            # More importantly, this is the fastest way to accomplish this.
            hint_reduced = typevar_to_hint.get(hint, hint)  # type: ignore[call-overload]
        # Else, a parent hint of this type variable mapped two or more type
        # variables. In this case, fallback to a slower and more complicated
        # approach that avoids worst-case edge cases. This includes recursion in
        # type variable mappings, which arises in non-trivial class hierarchies
        # involving two or more generics subscripted by two or more type
        # variables that circularly cycle between one another: e.g.,
        #     from typing import Generic
        #     class GenericRoot[T](Generic[T]): pass
        #
        #     # This directly maps {T: S}.
        #     class GenericLeaf[S](GenericStem[S]): pass
        #
        #     # This directly maps {S: T}, which then combines with the above
        #     # mapping to indirectly map {S: T, T: S}. Clearly, this indirect
        #     # mapping provokes infinite recursion unless explicitly handled.
        #     GenericLeaf[T]
        else:
            # Type hints previously reduced from this type variable, initialized
            # to this type variable.
            hint_reduced_prev = hint

            # Shallow copy of this type variable lookup table, coerced from an
            # immutable frozen dictionary into a mutable standard dictionary.
            # This enables type variables reduced by the iteration below to be
            # popped off this copy as a simple (but effective) recursion guard.
            typevar_to_hint_stack = typevar_to_hint.copy()

            #FIXME: [SPEED] *INEFFICIENT.* This has to be done either way, but
            #rather than reperform this O(n) algorithm on every single instance
            #of this type variable, this algorithm should simply be performed
            #exactly *ONCE* in the
            #reduce_hint_pep484_subbed_typevars_to_hints() reducer. Please
            #refactor this iteration over there *AFTER* the dust settles here.
            #FIXME: Actually, it's unclear how exactly this could be refactored
            #into the reduce_hint_pep484_subbed_typevars_to_hints()
            #reducer. This reduction here only searches for a single typevar in
            #O(n) time. Refactoring this over to
            #reduce_hint_pep484_subbed_typevars_to_hints() would require
            #generalizing this into an O(n**2) algorithm there, probably. Yow!

            # While...
            while (
                # This stack still contains one or more type variables that have
                # yet to be reduced by this iteration *AND*...
                typevar_to_hint_stack and
                # This hint is still a type variable...
                isinstance(hint_reduced, TypeVar)
            ):
                # Hint mapped to by this type variable if one or more parent
                # hints previously mapped this type variable to a hint *OR* this
                # hint as is (i.e., if this type variable is unmapped).
                #
                # Note that this one-liner destructively pops this type variable
                # off this temporary stack to prevent this type variable from
                # being reduced more than once by an otherwise recursive
                # mapping. Since this stack is local to this reducer, this
                # behaviour is only locally destructive and thus safe.
                hint_reduced = typevar_to_hint_stack.pop(
                    hint_reduced_prev, hint_reduced_prev)  # pyright: ignore

                # If this type variable maps to itself, this mapping is both
                # ignorable *AND* terminates this reduction.
                if hint_reduced is hint_reduced_prev:
                    break
                # Else, this type variable does *NOT* map to itself.

                # Map this type variable to this hint.
                hint_reduced_prev = hint_reduced
        # print(f'...to hint {hint} via type variable lookup table!')
    # Else, this type variable is unmapped.

    # ....................{ PHASE ~ 1 : bounds             }....................
    # If this hint is still a type variable (e.g., due to either not being
    # mapped by this lookup table *OR* being mapped by this lookup table to yet
    # another type variable)...
    if isinstance(hint_reduced, TypeVar):
        # PEP-compliant hint synthesized from all bounded constraints
        # parametrizing this type variable if any *OR* "None" otherwise (i.e.,
        # if this type variable is both unbounded *AND* unconstrained).
        #
        # Note this call is passed positional parameters due to memoization.
        hint_reduced = get_hint_pep484_typevar_bound_or_none(
            hint_reduced, exception_prefix)  # pyright: ignore

        # If this type variable is both unbounded *AND* unconstrained, this type
        # variable is currently *NOT* type-checkable and is thus ignorable.
        # Reduce this type variable to the ignorable singleton.
        if hint_reduced is None:
            return HINT_SANE_IGNORABLE
        # Else, this type variable is either bounded *OR* constrained. In either
        # case, preserve this newly synthesized hint.
        # print(f'Reducing PEP 484 type variable {repr(hint)} to {repr(hint_bound)}...')
        # print(f'Reducing non-beartype PEP 593 type hint {repr(hint)}...')
    # Else, one or more transitive parent hints previously mapped this type
    # variable to another hint.

    # ....................{ PHASE ~ 2                      }....................
    # Decide the recursion guard protecting this possibly recursive type
    # variable against infinite recursion. Note that:
    # * This guard intentionally applies to the original unreduced type variable
    #   (rather rather than the newly reduced hint decided by the prior phase).
    #   Thus, we pass "hint_recursable=hint" rather than
    #   "hint_recursable=hint_reduced".
    hint_sane = make_hint_sane_recursable(
        # The recursable form of this type alias is the original subscripted
        # hint tested above by the is_hint_recursive() recursion guard.
        hint_recursable=hint,
        # The non-recursable form of this type alias is the unsubscripted hint
        # underlying the original subscripted hint.
        hint_nonrecursable=hint_reduced,
        hint_parent_sane=hint_parent_sane,
    )

    # ....................{ RETURN                         }....................
    # Return this metadata.
    return hint_sane


def reduce_hint_pep484_subbed_typevars_to_hints(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    hint_parent_sane: Optional[HintSane] = None,
    exception_prefix: str = '',
) -> HintOrSane:
    '''
    Reduce the passed :pep:`484`-compliant **subscripted hint** (i.e., object
    produced by subscripting an unsubscripted hint originally parametrized by
    one or more :pep:`484`-compliant type variables with one or more child
    hints) to that unsubscripted hint and corresponding **type variable lookup
    table** (i.e., immutable dictionary mapping from those same type variables
    to those same child hints).

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Caveats
    -------
    This reducer does *not* validate these type variables to actually be type
    variables. Instead, this function defers that validation to the caller. Why?
    Efficiency, mostly. Avoiding the need to explicitly validate these type
    variables reduces the underlying mapping operation to a fast one-liner.

    This reducer validates the sizes of these two tuples to be constrained as:

    .. code-block:: python

       len(typevars) >= len(hints) > 0

    Informally, the passed hint *must* be subscripted by at least one child
    hint. For each such child hint, the unsubscripted hint originating this
    subscripted hint *must* be parametrized by a corresponding type variable.
    The converse is *not* the case, as:

    * For the first type variable, there also *must* exist a corresponding child
      hint to map to that type variable.
    * For *all* type variables following the first, there need *not* exist a
      corresponding child hint to map to that type variable. Type variables with
      *no* corresponding child hints are simply silently ignored (i.e.,
      preserved as type variables rather than mapped to other type hints).

    Equivalently:

    * Both of these tuples *must* be **non-empty** (i.e., contain one or more
      items).
    * This tuple of type variables *must* contain at least as many items as this
      tuple of child hints. Therefore:

      * This tuple of type variables *may* contain exactly as many items as this
        tuple of child hints.
      * This tuple of type variables *may* contain strictly more items than this
        tuple of child hints.
      * This tuple of type variables must *not* contain fewer items than this
        tuple of child hints.

    Parameters
    ----------
    hint : Hint
        Subscripted hint to be inspected.
    hint_parent_sane : Optional[HintSane]
        Either:

        * If the passed hint is a **root** (i.e., top-most parent hint of a tree
          of child hints), :data:`None`.
        * Else, the passed hint is a **child** of some parent hint. In this
          case, the **sanified parent type hint metadata** (i.e., immutable and
          thus hashable object encapsulating *all* metadata previously returned
          by :mod:`beartype._check.convert.convsanify` sanifiers after
          sanitizing the possibly PEP-noncompliant parent hint of this child
          hint into a fully PEP-compliant parent hint).

        Defaults to :data:`None`.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    HintOrSane
        Either:

        * If the unsubscripted hint (e.g., :class:`typing.Generic`) originating
          this subscripted hint (e.g., ``typing.Generic[S, T]``) is
          unparametrized by type variables, that unsubscripted hint as is.
        * Else, that unsubscripted hint is parametrized by one or more type
          variables. In this case, the **sanified type hint metadata** (i.e.,
          :class:`.HintSane` object) describing this reduction.

    Raises
    ------
    exception_cls
        If this type hint is unsubscripted.
    BeartypeDecorHintPep484TypeVarViolation
        If one of these type hints violates the bounds or constraints of one of
        these type variables.
    '''

    # ....................{ LOCALS                         }....................
    # Unsubscripted type alias originating this subscripted hint.
    hint_unsubbed = get_hint_pep_origin(
        hint=hint,
        exception_cls=BeartypeDecorHintPep484TypeVarException,
        exception_prefix=exception_prefix,
    )

    # Tuple of all type variables parametrizing this unsubscripted hint.
    #
    # Note that PEP 695-compliant "type" alias syntax superficially appears to
    # erroneously permit type aliases to be parametrized by non-type variables.
    # In truth, "type" syntax simply permits type aliases to be parametrized by
    # type variables that ambiguously share the same names as builtin types --
    # which then silently shadow those types for the duration of those aliases:
    #     >>> type muh_alias[int] = float | complex  # <-- *gulp*
    #     >>> muh_alias.__parameters__
    #     (int,)  # <-- doesn't look good so far
    #     >>> muh_alias.__parameters__[0] is int
    #     False  # <-- something good finally happened
    hint_unsubbed_typevars = get_hint_pep_typevars(hint_unsubbed)

    # Tuple of all child hints subscripting this subscripted hint.
    hint_childs = get_hint_pep_args(hint)
    # print(f'hint_childs: {repr(hint_childs)}')

    # ....................{ REDUCE                         }....................
    # Decide the type variable lookup table for this hint. Specifically, reduce
    # this subscripted hint to:
    # * The semantically useful unsubscripted hint originating this
    #   semantically useless subscripted hint.
    # * The type variable lookup table mapping all type variables parametrizing
    #   this unsubscripted hint to all non-type variable hints subscripting
    #   this subscripted hint.

    # ....................{ REDUCE ~ noop                  }....................
    # If either...
    if (
        # This unsubscripted hint is parametrized by *NO* type variables, *NO*
        # type variable lookup table can be produced by this reduction.
        #
        # Note this this is an uncommon edge case. Examples include:
        # * Parametrizations of the PEP 484-compliant "typing.Generic"
        #   superclass (e.g., "typing.Generic[S, T]"). In this case, the
        #   original unsubscripted "typing.Generic" superclass remains
        #   unparametrized despite that superclass later being parametrized.
        not hint_unsubbed_typevars or
        # This unsubscripted hint is parametrized by the exact same type
        # variables as this subscripted hint is subscripted by, in which case
        # the resulting type variable lookup table would uselessly be the
        # identity mapping from each of these type variables to itself. While an
        # identity type variable lookup table could trivially be produced, doing
        # so would convey *NO* meaningful semantics and thus be pointless.
        hint_childs == hint_unsubbed_typevars
    # Then reduce this subscripted hint to simply this unsubscripted hint, as
    # type variable lookup tables are then irrelevant.
    ):
        return hint_unsubbed
    # Else, this unsubscripted hint is parametrized by one or more type
    # variables. In this case, produce a type variable lookup table mapping
    # these type variables to child hints subscripting this subscripted hint.

    # ....................{ REDUCE ~ map                   }....................
    # Type variable lookup table mapping from each of these type variables to
    # each of these corresponding child hints.
    #
    # Note that we pass parameters positionally due to memoization.
    typevar_to_hint = _get_hint_pep484_typevars_to_hints(
        hint, hint_unsubbed_typevars, hint_childs, exception_prefix)
    # print(f'Mapped hint {hint} to type variable lookup table {typevar_to_hint}!')

    # ....................{ REDUCE ~ composite             }....................
    # Sanified metadata to be returned.
    hint_sane: HintSane = None  # type: ignore[assignment]

    # If this hint has *NO* parent, this is a root hint. In this case...
    if hint_parent_sane is None:
        # Metadata encapsulating this hint and type variable lookup table.
        hint_sane = HintSane(
            hint=hint_unsubbed, typevar_to_hint=typevar_to_hint)
    # Else, this hint has a parent. In this case...
    else:
        # If the parent hint is also associated with a type variable lookup
        # table...
        if hint_parent_sane.typevar_to_hint:
            # Full type variable lookup table merging the table associated this
            # parent hint with the table just decided above for this child hint,
            # efficiently defined as...
            typevar_to_hint = (
                # The type variable lookup table describing all transitive
                # parent hints of this hint with...
                hint_parent_sane.typevar_to_hint |  # type: ignore[operator]
                # The type variable lookup table describing this hint.
                #
                # Note that this table is intentionally the second rather than
                # first operand of this "|" operation, efficiently ensuring that
                # type variables mapped by this hint take precedence over type
                # variables mapped by transitive parent hints of this hint.
                typevar_to_hint
            )
        # Else, the parent hint is associated with *NO* such table.

        # Metadata encapsulating this hint and type variable lookup table, while
        # "cascading" any other metadata associated with this parent hint (e.g.,
        # recursable hint IDs) down onto this child hint as well.
        hint_sane = hint_parent_sane.permute_sane(
            hint=hint_unsubbed, typevar_to_hint=typevar_to_hint)

    # ....................{ RETURN                         }....................
    # print(f'Reduced subscripted hint {repr(hint)} to unsubscripted hint metadata {repr(hint_sane)}.')

    # Return this metadata.
    return hint_sane

# ....................{ PRIVATE ~ constants                }....................
_HINT_PEP484_TYPEVAR_RECURSABLE_DEPTH_MAX = 0
'''
Value of the optional ``hint_recursable_depth_max`` parameter passed to the
:func:`.is_hint_recursive` tester by the :func:`.reduce_hint_pep484_typevar`
reducer.

This depth ensures that :pep:`484`-compliant type variables are considered to be
recursive *only* after having been recursed into at most this many times before
(i.e., *only* after having been visited exactly once as a child hint of an
arbitrary parent hint). By definition, :pep:`484`-compliant type variables are
guaranteed to *never* be parent hints. Ergo, recursing into type variables
exactly once suffices to expand exactly one nesting level of non-trivial
recursive data structures. This mirrors the restrained recursion allowed by
other reducers.

Consider :pep:`695`-compliant type aliases, for example. Unlike type variables,
type aliases are guaranteed to *always* be parent hints. The value of the
optional ``hint_recursable_depth_max`` parameter passed to the
:func:`.is_hint_recursive` tester by :pep:`695`-specific reducers is thus one
greater than the value of this global. Nonetheless, the real-world effect is
exactly the same: exactly one nesting level of type aliases is expanded.

Consider the following :pep:`484`-compliant generic recursively subscripted by
itself via a :pep:`484`-compliant type variable ``T``:

.. code-block:: python

   class GenericList[T](): ...
   RecursiveGenericList = GenericList[GenericList]

Instances of this generic satisfying the type hint ``RecursiveGenericList``
contain an arbitrary number of other instances of this generic, exhibiting this
internal structure:

.. code-block:: python

   recursive_generic_list = GenericList()
   recursive_generic_list.append(recursive_generic_list)

Halting recursion at the first expansion of the type variable ``T`` then reduces
the type hint ``RecursiveGenericList`` to
``GenericList[GenericList[T]]``, , which reduces to
``GenericList[GenericList[HINT_SANE_RECURSIVE]]``, , which reduces to
``GenericList[GenericList]`` -- conveying exactly one layer of the internal
semantics of this recursive data structure.
'''

# ....................{ PRIVATE ~ getters                  }....................
@callable_cached
def _get_hint_pep484_typevars_to_hints(
    # Mandatory parameters.
    hint: Hint,
    hint_child_typevars: TupleTypeVars,
    hint_child_hints: TupleHints,

    # Optional parameters.
    exception_prefix: str = '',
) -> TypeVarToHint:
    '''
    Type variable lookup table mapping from the passed :pep:`484`-compliant
    **type variables** (i.e., :class:`.TypeVar` objects) to the associated
    passed type hints as key-value pairs of this table.

    This getter is memoized for efficiency. Notably, this getter creates and
    returns a dictionary mapping each type variable in the passed tuple of type
    variables to the associated type hint in the passed tuple of type hints with
    the same 0-based tuple index as that type variable.

    Parameters
    ----------
    hint: Hint
        Parent hint presumably both subscripted by these child hints. This
        parent hint is currently only used to generate human-readable exception
        messages in the event of fatal errors.
    hint_child_typevars : TupleTypeVars
        Tuple of one or more child type variables originally subscripting the
        origin underlying this parent hint.
    hint_child_hints : TupleHints
        Tuple of one or more child type hints subscripting this parent hint,
        which those type variables map to.
    exception_prefix: str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    TypeVarToHint
        Type variable lookup table mapping these type variables to hint_child_hints.

    Raises
    ------
    BeartypeDecorHintPep484TypeVarException
        If either:

        * This tuple of type variables is empty.
        * This tuple of type hints is empty.
        * This tuple of type hints contains more items than this tuple of type
          variables.
    BeartypeDecorHintPep484TypeVarViolation
        If one of these type hints violates the bounds or constraints of one of
        these type variables.
    '''
    assert isinstance(hint_child_typevars, tuple), (
        f'{repr(hint_child_typevars)} not tuple.')
    assert isinstance(hint_child_hints, tuple), (
        f'{repr(hint_child_hints)} not tuple.')
    assert isinstance(exception_prefix, str), (
        f'{repr(exception_prefix)} not string.')

    # ....................{ PREAMBLE                       }....................
    # If *NO* type variables were passed, raise an exception.
    if not hint_child_typevars:
        raise BeartypeDecorHintPep484TypeVarException(
            f'{exception_prefix}type hint {repr(hint)} '
            f'parametrized by no PEP 484 type variables.'
        )
    # Else, one or more type variables were passed.
    #
    # If *NO* type hint_child_hints were passed, raise an exception.
    elif not hint_child_hints:
        raise BeartypeDecorHintPep484TypeVarException(
            f'{exception_prefix}type hint {repr(hint)} '
            f'subscripted by no type hints.'
        )
    # Else, one or more type hint_child_hints were passed.
    #
    # If more type hints than type variables were passed, raise an exception.
    elif len(hint_child_hints) > len(hint_child_typevars):
        raise BeartypeDecorHintPep484TypeVarException(
            f'{exception_prefix}type hint {repr(hint)} '
            f'number of subscripted type hints {len(hint_child_hints)} exceeds '
            f'number of parametrized type variables {len(hint_child_typevars)} '
            f'(i.e., {len(hint_child_hints)} > {len(hint_child_typevars)}).'
        )
    # Else, either the same number of type hints and type variables were passed
    # *OR* more type variables than type hints were passed.

    # ....................{ MAP                            }....................
    # Type variable lookup table to be returned.
    typevar_to_hint: TypeVarToHint = {}

    #FIXME: Optimize into a "while" loop at some point. *sigh*
    # For each passed type variable and corresponding type hint...
    #
    # Note that:
    # * The C-based zip() builtin has been profiled to be the fastest means of
    #   iterating pairs in pure-Python, interestingly.
    # * If more type variables than type hints were passed, zip() silently
    #   ignores type variables with *NO* corresponding type hints -- exactly as
    #   required and documented by the above docstring.
    # print(f'Mapping hint_child_typevars {hint_child_typevars} -> hint_child_hints {hint_child_hints}...')
    for hint_child_typevar, hint_child_hint in zip(
        hint_child_typevars, hint_child_hints):
        # print(f'Mapping typevar {typevar} -> hint {hint}...')
        # print(f'is_hint_nonpep_type({hint})? {is_hint_nonpep_type(hint, False)}')

        # If this is *NOT* actually a type variable, raise an exception.
        #
        # Note that Python itself typically fails to validate this constraint,
        # thus requiring that we do so explicitly. For example:
        # * Ideally, *ALL* child hints parametrizing a PEP 695-compliant
        #   subscripted type alias would actually be type variables. Sadly, the
        #   "type" statement is excessively permissive under at least Python <=
        #   3.13 (and possibly newer Python releases as well): e.g.,
        #       >>> type muh_alias[int] = float | complex
        #       >>> muh_alias.__parameters__
        #       (int,)  # <-- pretty sure that's *NOT* a parameter, Python
        if not is_hint_pep484_typevar(hint_child_typevar):
            raise BeartypeDecorHintPep484TypeVarException(
                f'{exception_prefix}type hint {repr(hint)} '
                f'parametrization {repr(hint_child_typevar)} not '
                f'PEP 484 type variable (i.e., "typing.TypeVar" object).'
            )
        # Else, this is actually a type variable.

        #FIXME: Insufficient. Ideally, we would also validate this hint to be a
        #*SUBHINT* of this type variable. Specifically, if this type variable is
        #bounded by one or more bounded constraints, then we should validate
        #this hint to be a *SUBHINT* of those constraints: e.g.,
        #    class MuhClass(object): pass
        #
        #    # PEP 695 type alias parametrized by a type variable bound to a
        #    # subclass of the "MuhClass" type.
        #    type muh_alias[T: MuhClass] = T | int
        #
        #    # *INVALID.* Ideally, @beartype should reject this, as "int" is
        #    # *NOT* a subhint of "MuhClass".
        #    def muh_func(muh_arg: muh_alias[int]) -> None: pass
        #
        #Doing so is complicated, however, by forward reference proxies. For
        #obvious reasons, forward reference proxies are *NOT* safely resolvable
        #at this early decoration time that this function is typically called
        #at. If this hint either:
        #* Is itself a forward reference proxy, ignore rather than validate this
        #  hint as a subhint of these bounded constraints. Doing so is trivial
        #  by simply calling "is_beartype_forwardref(hint)" here.
        #* Is *NOT* itself a forward reference proxy but is transitively
        #  subscripted by one or more forward reference proxies, ignore rather
        #  than validate this hint as a subhint of these bounded constraints.
        #  Doing so is *EXTREMELY NON-TRIVIAL.* Indeed, there's *NO* reasonable
        #  way to do so directly here. Rather, we'd probably have to embrace an
        #  EAFP approach: that is, just crudely try to:
        #  * Detect whether this hint is a subhint of these bounded constraints.
        #  * If doing so raises an exception indicative of a forward reference
        #    issue, silently ignore that exception.
        #
        #  Of course, we're unclear what exception type that would even be. Does
        #  the beartype.door.is_subhint() tester even explicitly detect forward
        #  reference issues and raise an appropriate exception type? No idea.
        #  Probably *NOT*, honestly. Interestingly, is_subhint() currently even
        #  fails to support standard PEP 484-compliant forward references:
        #      >>> is_subhint('int', int)
        #      beartype.roar.BeartypeDoorNonpepException: Type hint 'int'
        #      currently unsupported by "beartype.door.TypeHint".
        #
        #Due to these (and probably more) issues, we currently *ONLY* validate
        #this hint to be a subhint of these bounded constraints...

        # If this hint is a PEP-noncompliant isinstanceable type (and thus *NOT*
        # an unresolvable forward reference proxy, which by definition is *NOT*
        # isinstanceable)...
        elif is_hint_nonpep_type(
            hint=hint_child_hint, is_forwardref_valid=False):
            # PEP-compliant type hint synthesized from all bounded constraints
            # parametrizing this type variable if any *OR* "None" otherwise.
            #
            # Note that this call is intentionally passed positional rather
            # positional keywords due to memoization.
            typevar_bound = get_hint_pep484_typevar_bound_or_none(
                hint_child_typevar, exception_prefix)
            # print(f'[{typevar}] is_object_issubclassable({typevar_bound})? ...')
            # print(f'{is_object_issubclassable(typevar_bound, False)}')

            # If...
            if (
                # This type variable was bounded or constrained *AND*...
                typevar_bound is not None and
                # These bounded constraints are PEP-noncompliant *AND*...
                #
                # PEP-compliant constraints are *NOT* safely passable to the
                # isinstance() or issubclass() testers, even if they technically
                # are isinstanceable or issubclassable. Why? Consider the
                # "typing.Any" singleton. Under newer Python versions, the
                # "typing.Any" singleton is actually defined as a subclassable
                # type. Although effectively *NO* real-world types subclass
                # "typing.Any", literally *ALL* objects (including types)
                # satisfy the "typing.Any" type hint. Passing "typing.Any" as
                # the second parameter to the issubclass() tester below would
                # thus erroneously reject (rather than silently accept) *ALL*
                # objects as unconditionally violating these bounds.
                not is_hint_pep(typevar_bound) and
                #FIXME: As follows, please:
                #* Unit test with "bound=Any". *sigh*
                #* Integration test with "sqlalchemy". *sigh*

                # These bounded constraints are issubclassable (i.e., an object
                # safely passable as the second parameter to the issubclass()
                # builtin) *AND*...
                #
                # Note that this function is memoized and thus permits *ONLY*
                # positional parameters.
                is_object_issubclassable(
                    typevar_bound,
                    # Ignore unresolvable forward reference proxies (i.e.,
                    # beartype-specific objects referring to user-defined
                    # external types that have yet to be defined).
                    False,
                ) and
                # This PEP-noncompliant isinstanceable type hint is *NOT* a
                # subclass of these bounded constraints...
                not issubclass(hint_child_hint, typevar_bound)  # type: ignore[arg-type]
            ):
                # Raise a type-checking violation.
                raise BeartypeDecorHintPep484TypeVarViolation(
                    message=(
                        f'{exception_prefix}type hint {repr(hint)} '
                        f'originally parametrized by '
                        f'PEP 484 type variable {repr(hint_child_typevar)} '
                        f'subsequently subscripted by '
                        f'child type hint {repr(hint_child_hint)} violating '
                        f"this type variable's bounds or constraints "
                        f'{repr(typevar_bound)}.'
                    ),
                    culprits=(hint_child_hint,),
                )
            # Else, this type variable was either:
            # * Unbounded and unconstrained.
            # * Bounded or constrained by a hint that is *NOT* issubclassable.
            # * Bounded or constrained by an issubclassable object that is the
            #   superclass of this corresponding hint, which thus satisfies
            #   these bounded constraints.
        # Else, this hint is *NOT* a PEP-noncompliant isinstanceable type.

        # Map this type variable to this hint with an optimally efficient
        # one-liner, silently overwriting any prior such mapping of this type
        # variable by either this call or a prior call of this function.
        typevar_to_hint[hint_child_typevar] = hint_child_hint

    # ....................{ RETURN                         }....................
    # Return this table, coerced into an immutable frozen dictionary.
    return FrozenDict(typevar_to_hint)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype exception raiser utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.error.utilerrraise` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_reraise_exception_placeholder() -> None:
    '''
    Test the
    :func:`beartype._util.error.utilerrraise.reraise_exception_placeholder`
    function.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.cache.utilcachecall import callable_cached
    from beartype._util.error.utilerrraise import reraise_exception_placeholder
    from pytest import raises
    from random import getrandbits

    # ....................{ CLASSES                        }....................
    class CachedException(ValueError):
        '''
        Test-specific exception intended to be raised below with messages
        containing placeholder substrings replaced by the
        :func:`.reraise_exception_placeholder` re-raiser.
        '''

        pass

    # ....................{ LOCALS                         }....................
    # Source substring to be hard-coded into the messages of all exceptions
    # raised by the low-level memoized callable defined below.
    TEST_SOURCE_STR = '{its_got_bro}'

    # ....................{ CALLABLES                      }....................
    @callable_cached
    def portend_low_level_winter(is_winter_coming: bool) -> str:
        '''
        Low-level memoized callable raising unreadable exceptions conditionally
        depending on the value of the passed parameter.
        '''

        if is_winter_coming:
            raise CachedException(
                f'{TEST_SOURCE_STR} intimates that winter is coming.')
        else:
            return 'PRAISE THE SUN'


    def portend_high_level_winter() -> None:
        '''
        High-level non-memoized callable calling the low-level memoized callable
        and reraising unreadable exceptions raised by the latter with equivalent
        readable exceptions.
        '''

        try:
            # Call the low-level memoized callable without raising exceptions.
            print(portend_low_level_winter(False))

            # Call the low-level memoized callable with raising exceptions.
            print(portend_low_level_winter(True))
        except CachedException as exception:
            # print('exception.args: {!r} ({!r})'.format(exception.args, type(exception.args)))
            reraise_exception_placeholder(
                exception=exception,
                source_str=TEST_SOURCE_STR,
                target_str=(
                    'Random "Song of Fire and Ice" spoiler' if getrandbits(1) else
                    'Random "Dark Souls" plaintext meme'
                ),
            )

    # ....................{ ASSERT                         }....................
    # Assert this high-level non-memoized callable raises the same type of
    # exception raised by this low-level memoized callable and preserve this
    # exception for subsequent assertion.
    with raises(CachedException) as exception_info:
        portend_high_level_winter()

    # Assert that this exception message does *NOT* contain the unreadable
    # source substring hard-coded into the messages of all exceptions raised by
    # this low-level memoized callable.
    assert TEST_SOURCE_STR not in str(exception_info.value)

    # Assert that exception messages may also contain *NO* source substrings.
    try:
        raise CachedException(
            "What's bravery without a dash of recklessness?")
    except Exception as exception:
        with raises(CachedException):
            reraise_exception_placeholder(
                exception=exception,
                source_str=TEST_SOURCE_STR,
                target_str='and man sees not light,',
            )

    # Assert that exception messages may be empty.
    try:
        raise CachedException()
    except Exception as exception:
        with raises(CachedException):
            reraise_exception_placeholder(
                exception=exception,
                source_str=TEST_SOURCE_STR,
                target_str='but only endless nights.',
            )

    # Assert that exception messages need *NOT* be strings.
    try:
        raise CachedException(0xDEADBEEF)
    except Exception as exception:
        with raises(CachedException):
            reraise_exception_placeholder(
                exception=exception,
                source_str=TEST_SOURCE_STR,
                target_str='Rise if you would, for that is our curse.',
            )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **configuration enumerations** (i.e., public enumerations whose members
may be passed as initialization-time parameters to the
:meth:`beartype._conf.confmain.BeartypeConf.__init__` constructor to configure
:mod:`beartype` with optional runtime type-checking behaviours).

Most of the public attributes defined by this private submodule are explicitly
exported to external users in our top-level :mod:`beartype.__init__` submodule.
This private submodule is *not* intended for direct importation by downstream
callers.
'''

# ....................{ IMPORTS                            }....................
from enum import (
    Enum,
    IntEnum,
    auto as next_enum_member_value,
    unique as die_unless_enum_member_values_unique,
)

# ....................{ ENUMERATIONS                       }....................
@die_unless_enum_member_values_unique
class BeartypeDecorationPosition(Enum):
    '''
    Enumeration of all kinds of **import hook decorator positions** (i.e.,
    competing locations to which the :func:`beartype.beartype` decorator will be
    implicitly injected into existing chains of one or more decorators
    decorating classes and callables defined by modules imported under
    :mod:`beartype.claw` import hooks, each with concomitant tradeoffs with
    respect to decorator interoperability and quality assurance).

    Attributes
    ----------
    FIRST : EnumMemberType
        **First (i.e., bottom-most) decorator position**, configuring
        :mod:`beartype.claw` import hooks to inject the
        :func:`beartype.beartype` decorator as the first (i.e., bottom-most)
        decorator in relevant decorator chains.

        Note that this position is *not* the default (and thus need be
        explicitly passed anywhere). Why? Various hand-wavy reasons, the most
        compelling of which is that this position ignores standard decorators
        and **thus violates PEP standards.** Notably, this position ignores:

        * The :pep:`484`-compliant :func:`typing.no_type_check` decorator, which
          then erroneously instructs :func:`beartye.beartype` to type-check
          classes and callables that should *not* be type-checked.
        * The :pep:`557`-compliant :func:`dataclasses.dataclass` decorator,
          which then prevents :func:`beartye.beartype` from type-checking
          dataclasses.
        * Explicitly configured :func:`beartype.beartype` decorators (e.g.,
          ``@beartype(conf=BeartypeConf(...))``), which then instructs
          :func:`beartye.beartype` to type-check classes and callables under
          differing configurations.

        When this position is used, implicit :func:`beartype.beartype`
        decorators injected by :mod:`beartype.claw` import hooks assume
        precedence over *all* other decorators (including those listed above),
        with predictably catastrophic results. Since this is almost never what
        anyone wants, this is *not* the default: e.g.,

        .. code-block:: python

           # Registering this import hook...
           from beartype import beartype, BeartypeConf, BeartypeDecorationPosition
           from beartype.claw import beartype_this_package
           beartype_this_package(conf=BeartypeConf(
               claw_decoration_position_types=BeartypeDecorationPosition.FIRST))

           # ...transforms chains of class decorators like this...
           from dataclasses import dataclass
           @beartype(conf=BeartypeConf(is_debug=True))
           @dataclass
           class ClassyData(object):
               integral_datum: int

           # ...into chains of class decorators like this.
           from dataclasses import dataclass
           @beartype(conf=BeartypeConf(is_debug=True))  # <-- *IGNORED*
           @dataclass  # <-- *IGNORED* by the @beartype decorator injected below
           @beartype   # <-- @beartype now ignores all of the above decorators!!
           class ClassyData(object):
               integral_datum: int

        In the above example, the default :func:`beartype.beartype` decorator
        injected by the :func:`beartype.claw.beartype_this_package` silently
        fails to type-check the :func:`dataclasses.dataclass` decorator and then
        overwrites the ``@beartype(conf=BeartypeConf(is_debug=True))`` decorator
        manually configured by the author of that third-party package.
        Consequently, caveats apply to usage of this position:

        * This position should only be applied to codebases that avoid
          explicitly decorating classes and/or callables with standard
          decorators, including:

          * The :pep:`484`-compliant :func:`typing.no_type_check` decorator.
          * The :pep:`557`-compliant :func:`dataclasses.dataclass` decorator.
          * The :func:`beartype.beartype` decorator itself.

        * Equivalently, this position should only be applied to codebases that
          implicitly decorate *all* classes and callables with
          :mod:`beartype.claw` import hooks.
        * Equivalently, if a codebase explicitly decorates even a single class
          or callable with the :func:`typing.no_type_check`,
          :func:`dataclasses.dataclass`, or :func:`beartype.beartype`
          decorators, this position should *not* be used.
        * Consequently, this position should *not* be applied to other packages
          outside your direct control.
        * In particular, this position should *not* be applied to all packages
          with the :func:`beartype.claw.beartype_all` import hook: e.g.,

          .. code-block:: python

             # Never do this. Srsly. Never do this. Srsly! Read this and weep.
             from beartype import BeartypeConf, BeartypeDecorationPosition
             from beartype.claw import beartype_all
             beartype_all(conf=BeartypeConf(
                 claw_decoration_position_types=BeartypeDecorationPosition.FIRST))
    LAST : EnumMemberType
        **Last (i.e., top-most) decorator position**, configuring
        :mod:`beartype.claw` import hooks to inject the
        :func:`beartype.beartype` decorator as the last (i.e., top-most)
        decorator in relevant decorator chains.

        Note that this position is already the default (and thus need *not* be
        explicitly passed anywhere). Why? Various hand-wavy reasons, the most
        compelling of which is that this position respects explicitly configured
        :func:`beartype.beartype` decorations (e.g.,
        ``@beartype(conf=BeartypeConf(...))``). When this position is used,
        explicit :func:`beartype.beartype` decorations assume precedence over
        implicit :func:`beartype.beartype` decorations injected by
        :mod:`beartype.claw` import hooks. Since that is almost always what
        everyone wants, this is the default position: e.g.,

        .. code-block:: python

           # Registering this import hook...
           from beartype import beartype, BeartypeConf, BeartypeDecorationPosition
           from beartype.claw import beartype_this_package
           beartype_this_package(conf=BeartypeConf(
               claw_decoration_position_funcs=BeartypeDecorationPosition.LAST))

           # ...transforms chains of function decorators like this...
           from functools import cache
           @cache
           @beartype(conf=BeartypeConf(is_debug=True))
           def chad_func() -> int:
               return 42

           # ...into chains of function decorators like this.
           from functools import cache
           @beartype  # <-- @beartype decorates last rather than first! \\o/
           @cache
           @beartype(conf=BeartypeConf(is_debug=True))
           def chad_func() -> int:
               return 42

        In the above example, the default :func:`beartype.beartype` decorator
        injected by the :func:`beartype.claw.beartype_this_package` import hook
        is silently ignored in favour of the non-default
        ``@beartype(conf=BeartypeConf(is_debug=True))`` decorator manually
        configured by the author of that third-party package.
    '''

    FIRST = next_enum_member_value()
    LAST = next_enum_member_value()


@die_unless_enum_member_values_unique
class BeartypeStrategy(Enum):
    '''
    Enumeration of all kinds of **type-checking strategies** (i.e., competing
    procedures for type-checking objects passed to or returned from
    :func:`beartype.beartype`-decorated callables, each with concomitant
    tradeoffs with respect to runtime complexity and quality assurance).

    Strategies are intentionally named according to `conventional Big O
    notation <Big O_>`__ (e.g., :attr:`BeartypeStrategy.On` enables the
    ``O(n)`` strategy). Strategies are established per-decoration at the
    fine-grained level of callables decorated by the :func:`beartype.beartype`
    decorator by setting the :attr:`beartype.BeartypeConf.strategy` parameter of
    the :class:`beartype.BeartypeConf` object passed as the optional ``conf``
    parameter to that decorator.

    Strategies enforce their corresponding runtime complexities (e.g., ``O(n)``)
    across *all* type-checks performed for callables enabling those strategies.
    For example, a callable configured by the :attr:`BeartypeStrategy.On`
    strategy will exhibit linear ``O(n)`` complexity as its overhead for
    type-checking each nesting level of each container passed to and returned
    from that callable.

    .. _Big O:
       https://en.wikipedia.org/wiki/Big_O_notation

    Attributes
    ----------
    O0 : EnumMemberType
        **No-time strategy** (i.e, disabling type-checking for a decorated
        callable by reducing :func:`beartype.beartype` to the identity
        decorator for that callable). Although seemingly useless, this strategy
        enables users to selectively blacklist (prevent) callables from being
        type-checked by our as-yet-unimplemented import hook. When implemented,
        that hook will type-check all callables within a package or module
        *except* those callables explicitly decorated by this strategy.
    O1 : EnumMemberType
        **Constant-time strategy** (i.e., the default ``O(1)`` strategy,
        type-checking a single randomly selected item of each container). As the
        default, this strategy need *not* be explicitly enabled.
    Ologn : EnumMemberType
        **Logarithmic-time strategy** (i.e., the ``O(log n)`` strategy,
        type-checking a randomly selected number of items ``log(len(obj))`` of
        each container ``obj``). This strategy is **currently unimplemented.**
        (*To be implemented by a future beartype release.*)
    On : EnumMemberType
        **Linear-time strategy** (i.e., the ``O(n)`` strategy, type-checking
        *all* items of a container). This strategy is **currently
        unimplemented.** (*To be implemented by a future beartype release.*)
    '''

    O0 = next_enum_member_value()
    O1 = next_enum_member_value()
    Ologn = next_enum_member_value()
    On = next_enum_member_value()


@die_unless_enum_member_values_unique
class BeartypeViolationVerbosity(IntEnum):
    '''
    Enumeration of all kinds of **violation verbosities** (i.e., positive
    integers in the inclusive range ``[1, 5]`` governing the verbosity of
    exception messages raised by type-checking wrappers generated by the
    :func:`beartype.beartype` decorator when either receiving parameters *or*
    returning values violating their annotated type hints).

    Verbosities transparently reduce to integers and can thus be used wherever
    integers are used (e.g., ``BeartypeViolationVerbosity.DEFAULT + 1`` is the next
    level of verbosity beyond that of the default). Verbosities are established
    per-decoration at the fine-grained level of callables decorated by the
    :func:`beartype.beartype` decorator by setting the
    :attr:`beartype.BeartypeConf.violation_verbosity` parameter of the
    :class:`beartype.BeartypeConf` object passed as the optional ``conf``
    parameter to that decorator.

    Attributes
    ----------
    MINIMAL : EnumMemberType
        **Minimal verbosity,** intended for end users potentially lacking core
        expertise in Python.
    DEFAULT : EnumMemberType
        **Default verbosity,** intended for a general developer audience assumed
        to be fluent in Python.
    MAXIMAL : EnumMemberType
        **Maximum verbosity,** extending the default verbosity with additional
        contextual metadata intended for debugging violations. This includes:

        * A machine-readable representation of the beartype configuration under
          which the current violation occurred.
    '''

    MINIMAL = next_enum_member_value()
    DEFAULT = next_enum_member_value()
    MAXIMAL = next_enum_member_value()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`646` **data submodule.**

This submodule exercises :pep:`646` support for statement-level tuple and type
variable unpacking (e.g., ``*TypeVarTuple('Ts')``, ``*Tuple[int, str]``)
implemented in the :func:`beartype.beartype` decorator by declaring unit tests
exercising these aliases. For safety, these tests are intentionally isolated
from the main test suite. Notably, this low-level submodule implements the
higher-level ``test_decor_pep646()`` unit test in the main test suite.

Caveats
-------
**This submodule requires the active Python interpreter to target at least
Python 3.11.0.** If this is *not* the case, importing this submodule raises an
:exc:`SyntaxError` exception.
'''

# ....................{ TESTS ~ iterator                   }....................
def unit_test_decor_pep646() -> None:
    '''
    Test :pep:`646` support implemented in the :func:`beartype.beartype`
    decorator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeDecorHintPep646Exception
    from beartype.typing import (
        Generic,
        Tuple,
        TypeVarTuple,
        Unpack,
    )
    from pytest import raises

    # ....................{ LOCALs                         }....................
    # Arbitrary type variable tuple.
    Ts = TypeVarTuple('Ts')

    # ....................{ CLASSES                        }....................
    @beartype
    class ItsMotions(Generic[Unpack[Ts]]):
        '''
        Generic over an arbitrary number of type variables, decorated by
        :func:`beartype.beartype` and defined by subclassing the
        :class:`.Generic` superclass subscripted by the :pep:`646`-compliant
        :obj:`typing.Unpack` type hint factory subscripted by a
        :pep:`646`-compliant type variable tuple.

        Note that this is semantically equivalent to:

        .. code-block:: python

           class ItsMotions(Generic[*Ts]):
        '''

        pass


    # Arbitrary instance of this class.
    render_up_its_majesty = ItsMotions()

    # ....................{ CALLABLES                      }....................
    #FIXME: *LOL*. "Tuple[*Ts] == Tuple[Unpack[Ts]] == Tuple[object]" after
    #reduction, a fixed-length tuple hint. Clearly, however, "Tuple[Unpack[Ts]]"
    #should instead be semantically equivalent to a variadic tuple hint: e.g.,
    #    # This is what we want!
    #    Tuple[*Ts] == Tuple[Unpack[Ts]] == Tuple[object, ...]
    #
    #In other words, how exactly we reduce "Unpack[Ts]" conditionally depends on
    #the parent type hint containing this "Unpack[Ts]". Although we could begin
    #passing parent type hints around, that seems pretty crazy. Maybe? Or maybe
    #not. This reducer is uncached anyway. That said, consider the recursive
    #tree of all possible parent type hints. In the worst case, a child type
    #hint could conceivably need that. It might be saner, therefore, to... hmmm.
    #
    #What, exactly? The issue is that the get_hint_pep_sign_or_now() getter now
    #discriminates between fixed-length and variadic type hints! However,
    #deciding whether a tuple type hint is fixed-length or variadic is
    #increasingly non-trivial. Specifically:
    #
    #* A parent tuple hint is variadic if subscripted by either:
    #  * A single child type hint with sign "HintSignUnpack".
    #  * Exactly two child type hints, the last of which is an ellipsis.
    #* Else, that hint is fixed-length.
    #
    #That takes care of detection. Pretty insane, but certainly feasible.
    #
    #We still need to handle "Tuple[Unpack[Ts]] == Tuple[object, ...]"
    #reduction, however. Or maybe not? Maybe everything will magically work!?
    #FIXME: Actually, even detecting is insufficient. In truth, there no longer
    #exists a hard distinction between "fixed-length" and "variadic" tuple
    #hints. A tuple hint may now contain effectively arbitrary combinations
    #of either form by unpacking tuple hints inside other tuple hints.
    #
    #Sadly, this suggests that:
    #* We'll want to unwind our prior detection of:
    #  * Fixed-length tuple hints as "HintSignTupleFixed".
    #  * Variadic tuple hints as "HintSignTuple".
    #* Instead, handle both fixed-length and variadic tuple hints under the same
    #  "HintSignTuple" sign. *WOOPS.*
    #* Remove the "HintSignTupleFixed" sign, which is no longer required.
    #* Handle tuple hint unpacking inside tuple hints.
    @beartype
    def of_all_the_grace(
        and_to_the_damp_leaves: ItsMotions,
        *and_beauty_that_endued: *Ts
    # ) -> Tuple[ItsMotions, *Ts]:
    ) -> object:
        '''
        Arbitrary callable simply returning the tuple of all passed variadic
        positional parameters prepended by the passed generic, decorated by
        :func:`beartype.beartype` and correctly annotated by:

        * A variadic positional parameter typed as a :pep:`646`-compliant
          unpacked type variable tuple hint.
        * A return typed as a :pep:`484`-compliant tuple type hint subscripted
          by the same :pep:`646`-compliant unpacked type variable tuple hint.
        '''

        # Return this variadic positional parameter prepended by this generic.
        return (and_to_the_damp_leaves,) + and_beauty_that_endued

    # ....................{ PASS                           }....................
    # Assert that this callable returns a tuple of all passed positional
    # parameters.
    assert of_all_the_grace(
        render_up_its_majesty, 'Of all the grace', 'and beauty', 'that endued') == (
        render_up_its_majesty, 'Of all the grace', 'and beauty', 'that endued')

    # ....................{ FAIL                           }....................
    # Assert that @beartype raises the expected exception when decorating a
    # callable annotated by a hint unpacking a PEP 646-noncompliant object
    # (i.e., *ANY* object other than a PEP 646-compliant type variable tuple).
    with raises(BeartypeDecorHintPep646Exception):
        @beartype
        def scatter_its_music(
            *on_the_unfeeling_storm: Unpack[Tuple[str]]) -> str:
            return on_the_unfeeling_storm[0]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **PEP-agnostic type hint reduction** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._check.convert._reduce.redhint` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytmark import ignore_warnings

# ....................{ TESTS ~ reducers                   }....................
def test_reduce_hint() -> None:
    '''
    Test the private
    :func:`beartype._check.convert._reduce.redhint.reduce_hint` reducer.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintNonpepNumpyException
    from beartype.typing import (
        Annotated,
    )
    from beartype.vale import IsEqual
    from beartype._cave._cavefast import NoneType
    from beartype._check.convert._reduce.redhint import reduce_hint
    from beartype._check.metadata.hint.hintsane import HintSane
    from beartype._conf.confmain import BeartypeConf
    from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
    from beartype._data.cls.datacls import TYPES_PEP484_GENERIC_IO
    from beartype._data.hint.datahinttyping import (
        Pep484TowerComplex,
        Pep484TowerFloat,
    )
    from beartype._data.hint.pep.sign.datapepsigns import HintSignAnnotated
    from beartype._util.hint.pep.proposal.pep593 import is_hint_pep593
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign
    # from beartype_test.a00_unit.data.hint.pep.proposal.data_pep484 import (
    #     T_int,
    #     T_str_or_bytes,
    # )
    from beartype_test._util.module.pytmodtest import is_package_numpy
    from dataclasses import InitVar
    from pytest import raises

    # Intentionally import PEP 484-specific type hint factories.
    from typing import Protocol

    # ..................{ LOCALS                             }..................
    # Keyword arguments to be passed to all calls to reduce_hints() below.
    kwargs = {
        'conf': BEARTYPE_CONF_DEFAULT,
    }

    # ..................{ CORE                               }..................
    # Assert this reducer preserves an isinstanceable type as is.
    assert reduce_hint(hint=int, **kwargs) == HintSane(int)

    # Assert this reducer reduces "None" to "type(None)".
    assert reduce_hint(hint=None, **kwargs) == HintSane(NoneType)

    # ..................{ PEP 484 ~ tower                    }..................
    # Assert this reducer expands the builtin "float" and "complex" types to
    # their corresponding numeric towers when configured to do so.
    hint_pep484_tower_float_sane = reduce_hint(
        hint=float,
        conf=BeartypeConf(is_pep484_tower=True),
    )
    hint_pep484_tower_complex_sane = reduce_hint(
        hint=complex,
        conf=BeartypeConf(is_pep484_tower=True),
    )
    assert hint_pep484_tower_float_sane.hint == Pep484TowerFloat
    assert hint_pep484_tower_complex_sane.hint == Pep484TowerComplex

    # Assert this reducer preserves the builtin "float" and "complex" types as
    # is when configured to disable the implicit numeric tower.
    assert reduce_hint(
        hint=float,
        conf=BeartypeConf(is_pep484_tower=False),
    ) == HintSane(float)
    assert reduce_hint(
        hint=complex,
        conf=BeartypeConf(is_pep484_tower=False),
    ) == HintSane(complex)

    #FIXME: Excise us up, please. We no longer reduce type variables. *sigh*
    #FIXME: *WAIT*. We absolutely do reduce type variables. Resurrect this!
    # # ..................{ PEP 484 ~ typevar                  }..................
    # # Assert this reducer preserves unbounded type variables as is.
    # assert reduce_hint(hint=T, **kwargs) is T
    #
    # # Assert this reducer reduces bounded type variables to their upper bound.
    # assert reduce_hint(hint=T_int, **kwargs) is int
    #
    # # Union of all constraints parametrizing a constrained type variable,
    # # reduced from that type variable.
    # typevar_constraints_union = reduce_hint(hint=T_str_or_bytes, **kwargs)
    #
    # # Assert this union contains all constraints parametrizing this variable.
    # assert str   in typevar_constraints_union.__args__
    # assert bytes in typevar_constraints_union.__args__

    # ..................{ PEP 544                            }..................
    # For each PEP 484-compliant "typing" IO generic superclass...
    for pep484_generic_io in TYPES_PEP484_GENERIC_IO:
        # Metadata encapsulating the equivalent PEP 544-compliant protocol
        # reduced from this generic.
        pep544_protocol_io_sane = reduce_hint(hint=pep484_generic_io, **kwargs)

        # This protocol.
        pep544_protocol_io = pep544_protocol_io_sane.hint

        # Assert this protocol is either...
        assert (
            # A PEP 593-compliant type metahint generalizing a protocol *OR*...
            is_hint_pep593(pep544_protocol_io) or
            # A PEP 544-compliant protocol.
            issubclass(pep544_protocol_io, Protocol)
        )

    # ..................{ PEP 557                            }..................
    # Assert this reducer reduces an "InitVar" to its subscripted argument.
    assert reduce_hint(hint=InitVar[str], **kwargs) == HintSane(str)

    # ..................{ PEP 593                            }..................
    # Assert this reducer reduces a beartype-agnostic metahint to the
    # lower-level hint it annotates.
    assert reduce_hint(hint=Annotated[int, 42], **kwargs) == HintSane(int)

    # Assert this reducer preserves a beartype-specific metahint as is.
    leaves_when_laid = Annotated[str, IsEqual['In their noonday dreams.']]
    assert reduce_hint(hint=leaves_when_laid, **kwargs) == HintSane(
        leaves_when_laid)

    # ..................{ NUMPY                              }..................
    # If a recent version of NumPy is importable...
    if is_package_numpy():
        # Defer third party imports.
        from numpy import float64
        from numpy.typing import NDArray

        # Metadata encapsulating the reduction of a NumPy array type hint to a
        # beartype validator.
        ndarray_reduced_sane = reduce_hint(hint=NDArray[float64], **kwargs)

        # This beartype validator.
        ndarray_reduced = ndarray_reduced_sane.hint

        # Assert this validator is a "typing{_extensions}.Annotated" hint.
        assert get_hint_pep_sign(ndarray_reduced) is HintSignAnnotated

        # Assert that reducing a "numpy.typing.NDArray" type hint
        # erroneously subscripted by an object *NOT* reducible to a data
        # type raises the expected exception.
        with raises(BeartypeDecorHintNonpepNumpyException):
            reduce_hint(
                hint=NDArray['From_my_wings_are_shaken_the_dews_that_waken'],
                **kwargs
            )

# ....................{ TESTS ~ raiser                     }....................
# Prevent pytest from capturing and displaying all expected non-fatal
# beartype-specific warnings emitted by this test. Urgh!
@ignore_warnings(DeprecationWarning)
def test_reduce_hint_ignorable(hints_pep_meta, hints_ignorable) -> None:
    '''
    Test the private
    :func:`beartype._check.convert._reduce.redhint.reduce_hint` reducer with
    respect to ignorable hints.

    Parameters
    ----------
    hints_pep_meta : tuple[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        Tuple of type hint metadata describing sample type hints exercising edge
        cases in the :mod:`beartype` codebase.
    hints_ignorable : frozenset
        Frozen set of ignorable PEP-agnostic type hints.
    '''

    # Defer test-specific imports.
    from beartype._check.metadata.hint.hintsane import HINT_SANE_IGNORABLE
    from beartype._check.convert._reduce.redhint import reduce_hint
    from beartype_test.a00_unit.data.hint.data_hint import (
        HINTS_NONPEP_UNIGNORABLE)

    # Assert this tester accepts ignorable type hints.
    for hint_ignorable in hints_ignorable:
        assert reduce_hint(hint_ignorable) is HINT_SANE_IGNORABLE

    # Assert this tester rejects unignorable PEP-noncompliant type hints.
    for hint_unignorable in HINTS_NONPEP_UNIGNORABLE:
        assert reduce_hint(hint_unignorable) is not HINT_SANE_IGNORABLE

    # Assert this tester:
    # * Accepts unignorable PEP-compliant type hints.
    # * Rejects ignorable PEP-compliant type hints.
    for hint_pep_meta in hints_pep_meta:
        # True only if this hint reduces to the ignorable "Any" singleton.
        is_hint_ignorable = reduce_hint(hint_pep_meta.hint) is HINT_SANE_IGNORABLE

        # Assert this hint is either ignorable or unignorable as expected.
        assert hint_pep_meta.is_ignorable == is_hint_ignorable



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- and :pep:`604`-compliant **union type hint utility**
unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep484604` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ tester                     }....................
def test_is_hint_pep604() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep484604.is_hint_pep604` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.hint.pep.proposal.pep484604 import is_hint_pep604
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_10

    # If the active Python interpreter targets Python >= 3.10 and thus supports
    # PEP 604...
    if IS_PYTHON_AT_LEAST_3_10:
        # Assert this tester accepts a PEP 604-compliant union.
        assert is_hint_pep604(int | str | None) is True
    # Else, this interpreter targets Python < 3.10 and thus fails to support PEP
    # 604.

    # Assert this tester rejects an arbitrary PEP 604-noncompliant object.
    is_hint_pep604('Meet in the vale, and one majestic River,')

# ....................{ TESTS ~ factory                    }....................
def test_make_hint_pep484604_union() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep484604.make_hint_pep484604_union`
    factory.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep604Exception
    from beartype.typing import (
        List,
        Union,
    )
    from beartype._util.hint.pep.proposal.pep484604 import (
        make_hint_pep484604_union)
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_10
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert that this factory when passed a 1-tuple containing an arbitrary
    # child type hint simply returns that hint as is.
    assert make_hint_pep484604_union((int,)) is int

    # Assert that this factory when passed a tuple containing two or more PEP
    # 604-compliant child type hints returns either...
    assert make_hint_pep484604_union((int, List[str], None)) == (
        # If the active Python interpreter targets Python >= 3.10 and thus
        # supports PEP 604, a PEP 604-compliant new-style union of these hints;
        int | List[str] | None
        if IS_PYTHON_AT_LEAST_3_10 else
        # Else, the active Python interpreter targets Python < 3.10 and thus
        # fails to support PEP 604. In this case, a PEP 484-compliant old-style
        # union of these hints.
        Union[int, List[str], None]
    )

    # If the active Python interpreter targets Python >= 3.10 and thus supports
    # both PEP 593 (i.e., "typing.Annotated[...]" type hints) *AND* 604...
    if IS_PYTHON_AT_LEAST_3_10:
        # Defer version-specific imports.
        from beartype.typing import Annotated

        # Assert that this factory when passed a tuple containing two or more
        # child type hints at least one of which is PEP 593-compliant and thus
        # *NOT* PEP 604-compliant returns a PEP 484-compliant old-style union of
        # these hints.
        assert make_hint_pep484604_union(
            (bool, List[int], None, Annotated[bytes, None])) == Union[
             bool, List[int], None, Annotated[bytes, None]]
    # Else, the active Python interpreter targets Python < 3.10 and thus fails
    # to support at least PEP 604. In this case, don't even bother. Run away!

    # ....................{ FAIL                           }....................
    # Assert that this factory raises the expected exception when passed the
    # empty tuple.
    with raises(BeartypeDecorHintPep604Exception):
        make_hint_pep484604_union(())



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`484`-compliant **type variable reduction** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._check.convert._reduce._pep.pep484.redpep484typevar` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ reduce                     }....................
def test_reduce_hint_pep484_subbed_typevars_to_hints() -> None:
    '''
    Test the private
    :mod:`beartype._check.convert._reduce._pep.pep484.redpep484typevar.reduce_hint_pep484_subbed_typevars_to_hints`
    reducer.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep484TypeVarException
    from beartype.typing import (
        Generic,
    )
    from beartype._check.convert._reduce._pep.pep484.redpep484typevar import (
        reduce_hint_pep484_subbed_typevars_to_hints)
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_12
    from beartype._data.hint.datahinttyping import (
        S,
        T,
    )
    from pytest import raises

    # ....................{ PEP 484                        }....................
    # Assert that this reducer reduces the PEP 484-compliant "typing.Generic"
    # superclass subscripted by only type variables to simply that superclass.
    assert reduce_hint_pep484_subbed_typevars_to_hints(
        Generic[S, T]) is Generic

    # ....................{ PEP 695                        }....................
    # If the active Python interpreter targets Python >= 3.12 and thus supports
    # PEP 695...
    #
    # Note that PEP 695-compliant type aliases are used here as a reference
    # "subscripted hint", due to being the easiest subscripted hints to define,
    # use, and debug (under Python >= 3.12, anyway). Valid alternatives include:
    # * PEP 484- or 585-compliant subscripted generics, which are considerably
    #   more cumbersome to define.
    if IS_PYTHON_AT_LEAST_3_12:
        # Defer version-specific imports.
        from beartype_test.a00_unit.data.pep.pep695.data_pep695util import (
            unit_test_reduce_hint_pep484_subbed_typevars_to_hints)

        # Perform this test.
        unit_test_reduce_hint_pep484_subbed_typevars_to_hints()
    # Else, this interpreter targets Python < 3.12 and thus fails to support PEP
    # 695.

    # ....................{ FAIL                           }....................
    # Assert this reducer raises the expected exception when passed an object
    # that is *NOT* a PEP 695-compliant subscripted type alias.
    with raises(BeartypeDecorHintPep484TypeVarException):
        reduce_hint_pep484_subbed_typevars_to_hints(
            'In thy devastating omnipotence,')

# ....................{ TESTS ~ getter                     }....................
def test_get_hint_pep484_typevars_to_hints() -> None:
    '''
    Test the private
    :func:`beartype._check.convert._reduce._pep.pep484.redpep484typevar._get_hint_pep484_typevars_to_hints`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import (
        BeartypeDecorHintPep484TypeVarException,
        BeartypeDecorHintPep484TypeVarViolation,
    )
    from beartype._data.hint.datahinttyping import (
        S,
        T,
    )
    from beartype_test.a00_unit.data.hint.pep.proposal.data_pep484 import (
        T_any,
        T_sequence,
        T_str_or_bytes,
    )
    from beartype._check.convert._reduce._pep.pep484.redpep484typevar import (
        _get_hint_pep484_typevars_to_hints)
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # Type variable lookup table, initialized below.
    typevar_to_hint = {}

    # ....................{ PASS                           }....................
    # Assert that this mapper correctly maps a single type variable to a single
    # type hint.
    typevar_to_hint = _get_hint_pep484_typevars_to_hints(None, (S,), (int,))
    assert typevar_to_hint == {S: int}

    # Assert that this mapper correctly maps a single unbounded type variable
    # to a single type hint.
    typevar_to_hint = _get_hint_pep484_typevars_to_hints(
        None, (T_any,), (bool,))
    assert typevar_to_hint == {T_any: bool}

    # Assert that this mapper correctly maps a single type variable bounded to a
    # type to a single type hint satisfying that type.
    typevar_to_hint = _get_hint_pep484_typevars_to_hints(
        None, (T_sequence,), (list,))
    assert typevar_to_hint == {T_sequence: list}

    # Assert that this mapper correctly maps a single type variable constrained
    # to two or more types to a single type hint satisfying at least one of
    # those types.
    typevar_to_hint = _get_hint_pep484_typevars_to_hints(
        None, (T_str_or_bytes,), (bytes,))
    assert typevar_to_hint == {T_str_or_bytes: bytes}

    # Assert that this mapper correctly maps multiple type variables to multiple
    # type hints, including to a previously mapped type variable by silently
    # overwriting the type hint previously mapped to that type variable with the
    # corresponding passed hint.
    typevar_to_hint = _get_hint_pep484_typevars_to_hints(
        None, (T, S,), (float, complex,))
    assert typevar_to_hint == {S: complex, T: float}

    # ....................{ FAIL                           }....................
    # Assert that this mapper raises the expected exception when passed *NO*
    # type variables.
    with raises(BeartypeDecorHintPep484TypeVarException):
        _get_hint_pep484_typevars_to_hints(None, (), (bool,))

    # Assert that this mapper raises the expected exception when passed a type
    # variable that is *NOT* actually a type variables.
    with raises(BeartypeDecorHintPep484TypeVarException):
        _get_hint_pep484_typevars_to_hints(
            None,
            (S, 'His brother Death. A rare and regal prey',),
            (int, str,),
        )

    # Assert that this mapper raises the expected exception when passed *NO*
    # type hints.
    with raises(BeartypeDecorHintPep484TypeVarException):
        _get_hint_pep484_typevars_to_hints(None, (S,), ())

    # Assert that this mapper raises the expected exception when passed more
    # type hints than type variables.
    with raises(BeartypeDecorHintPep484TypeVarException):
        _get_hint_pep484_typevars_to_hints(
            None, (S, T,), (int, bool, complex,))

    # Assert that this mapper raises the expected violation when passed a type
    # hint violating the bounds of a passed type variable.
    with raises(BeartypeDecorHintPep484TypeVarViolation):
        _get_hint_pep484_typevars_to_hints(
            None, (S, T, T_sequence,), (float, complex, int,))

    #FIXME: Uncomment *AFTER* we generalize this mapper to type-check violations
    #against arbitrarily complex constraints.
    # # Assert that this mapper raises the expected violation when passed a type
    # # hint violating the constraints of a passed type variable.
    # with raises(BeartypeDecorHintPep484TypeVarViolation):
    #     map_pep484_typevars_to_hints(
    #         typevar_to_hint=typevar_to_hint,
    #         typevars=(S, T, T_str_or_bytes,),
    #         hints=(float, complex, ,),
    #     )
